'''A handler is responsible for handling a HTTP request'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from http import HTTPStatus
from http import client as hc
import socket, threading, urllib.parse

from appy.model.base import Base
from appy.server.guard import Guard
from appy.server.error import Error
from appy.server.static import Static
from appy.server.request import Request
from appy.model.utils import Object as O
from appy.server.response import Response
from appy.server.languages import Languages
from appy.server.traversal import Traversal
from appy.pod.actions import EvaluationError
from appy.utils import asDict, MessageException

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
TRANS_RETRY = 'Conflict occurred (%s). Replay transaction (#%d)...'
TRANS_ERR   = 'Unresolved conflict occurred (%s).'
SOCK_TO     = 'Client socket timeout: %s.'
RQ_LINE_KO  = 'Bad request line "%s".'
HTTP_V_KO   = 'Wrong HTTP version "%s".'
HTTP_V_UNS  = 'Unsupported HTTP version "%s".'
HEAD_TOO_LG = 'Line too long: %s...'
HEAD_T_MANY = 'Too many headers: %s...'
T_READ_SOCK = 'Reading request line on rfile...'
T_READ_GOT  = 'Got request line: %s (length: %d)'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class MethodsCache(dict):
    '''A dict of method results, used to avoid executing more than once the same
       method, while handling a request. Every handler implements such a
       cache.'''

    def call(self, o, method, class_=None, cache=True):
        '''Call p_method on some p_o(bject). m_method can be an instance method
           on p_o; it can also be a static method. In this latter case, p_o is
           the tool and the static method, defined in p_class_, will be called
           with the tool as unique arg.

           If the method result is already in the cache, it will simply be
           returned. Else, the method will be executed, its result will be
           stored in the cache and returned.

           If p_cache is False, caching is disabled and the method is always
           executed.
        '''
        # Disable the cache for lambda functions
        name = method.__name__
        cache = False if name == '<lambda>' else cache
        # Call the method if cache is not needed
        if not cache: return method(o)
        # If first arg of method is named "tool" instead of the traditional
        # "self", we cheat and will call the method with the tool as first arg.
        # This will allow to consider this method as if it was a static method
        # on the tool.
        cheat = False
        if not class_ and (method.__code__.co_varnames[0] == 'tool'):
            prefix = o.class_.name
            o = o.tool
            cheat = True
        # Build the unique key allowing to store method results in the cache.
        # The key is of the form
        #                    <object id>:<method name>
        # In the case of a static method, "object id" is replaced with the class
        # name.
        if not cheat:
            prefix = class_.__name__ if class_ else str(o.iid)
        key = '%s:%s' % (prefix, name)
        # Return the cached value if present in the method cache
        if key in self: return self[key]
        # No cached value: call the method, cache the result and return it
        r = method(o)
        self[key] = r
        return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Handler:
    '''Abstract handler'''

    # A handler handles an incoming HTTP request. Class "Handler" is the
    # abstract base class for all concrete handlers, as defined hereafter.
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #  HttpHandler   | Handles HTTP requests. In the Appy HTTP server, every
    #                | request is managed by a thread. At server startup, a
    #                | configurable number of threads are run and are waiting
    #                | for requests. Every time a request hits the server, it is
    #                | assigned to a thread, that instantiates a handler and
    #                | manages the request.
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # VirtualHandler | When the Appy HTTP server starts, it is like if he had to
    #                | handle a virtual request. For that special case, an for
    #                | other similar cases, like scheduler jobs, an instance of
    #                | VirtualHandler is created.
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # In the following static attribute, we maintain a registry of the currently
    # instantiated handlers. Every handler is keyed by the ID of the thread that
    # is running it.
    registry = {}

    # Make some names available here
    Static = Static

    # Every handler can use a special char allowing to easily identify entries
    # produced from it in log files.
    logChar = ''

    @classmethod
    def add(class_, handler):
        '''Adds a new p_handler into the registry'''
        id = threading.get_ident()
        if id in class_.registry: return
        class_.registry[id] = handler

    @classmethod
    def remove(class_):
        '''Remove p_handler from the registry'''
        del class_.registry[threading.get_ident()]

    @classmethod
    def get(class_):
        '''Returns the handler for the currently executing thread'''
        return class_.registry[threading.get_ident()]

    # Make the Language class available in the handler namespace
    Languages = Languages

    def init(self):
        '''Define attributes which are common to any handler'''
        # A lot of object methods are executed while handling a request. Dict
        # "method" hereafter caches method results: it avoids executing more
        # than once the same method. Only methods without args are cached, like
        # methods defined on field attributes such as "show".
        self.methods = MethodsCache()
        # Appy and apps may use the following object to cache any other element
        self.cache = O()
        # Create a guard, a transient object for managing security
        self.guard = Guard(self)

    def customInit(self):
        '''The app can define a method on its tool, named "initialiseHandler",
           allowing to perform custom initialisation on a freshly created
           handler (ie for caching some highly requested info).'''
        if not hasattr(self.tool, 'initialiseHandler'): return
        self.tool.initialiseHandler(self)

    def clientIP(self):
        '''Gets the IP address of the client'''
        # Check header key X-Forwarded-For first
        return self.headers.get('X-Forwarded-For') or self.clientHost

    # This dict allows, on a concrete handler, to find the data to log
    logAttributes = O(ip='self.clientIP()', port='str(self.clientPort)',
      command='self.command', path='self.path', protocol='self.requestVersion',
      message='message', user='self.guard.userLogin',
      agent='self.headers.get("User-Agent")')

    def log(self, type, level, message=None):
        '''Logs, in the logger determined by p_type, a p_message at some
           p_level, that can be "debug", "info", "warning", "error" or
           "critical". p_message can be empty: in this case, the log entry will
           only contain the predefined attributes as defined by the
           appy.database.log.Config.'''
        server = self.server
        # Add the "log char" to the message when relevant
        if message and self.logChar:
            message = '%s %s' % (self.logChar, message)
        logger = getattr(server.loggers, type)
        cfg = getattr(server.config.log, type)
        # Get the parts of the message to dump
        r = []
        for part in cfg.messageParts:
            try:
                value = eval(getattr(Handler.logAttributes, part))
                if value is not None:
                    r.append(value)
            except AttributeError:
                pass
        # Call the appropriate method on the logger object corresponding to the
        # log p_level.
        getattr(logger, level)(cfg.sep.join(r))

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Inject class Handler on class Base to avoid circular package dependencies
Base.Handler = Handler

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class HttpHandler(Handler):
    '''Handles incoming HTTP requests'''

    # This is a "real" request handler
    fake = False

    # Store, for every logged user, the date/time of its last access
    onlineUsers = {}

    # Supported HTTP methods
    supportedHttpMethods = asDict(['GET', 'POST'])

    def __init__(self, clientSocket, server):
        # The global config
        self.config = server.config
        # The client socket from which the client request will be read
        self.clientSocket = clientSocket
        self.clientHost, self.clientPort = clientSocket.getpeername()
        # A back reference to the Appy server
        self.server = server
        # Create a buffered file for reading data on the socket
        self.rfile = clientSocket.makefile('rb', -1)
        # Create a Response object, used to build the response to send back to
        # the client.
        self.resp = Response(self)
        # Must the client socket be closed after the request has been handled ?
        self.closeSocket = True # Until now
        # The debug level
        self.debugLevel = self.config.server.debugLevel

    def init(self):
        '''Additional initialization, only applicable for dynamic content'''
        # Update the response object with headers being specific to a dynamic
        # response.
        self.resp.initDynamic()
        # An instance of appy.ui.validate.Validator will be created here, when
        # data sent via the ui requires to be validated.
        self.validator = None
        # Search criteria may be cached
        self.criteria = None
        # Must we commit data into the database ?
        self.commit = False
        # Set here a link to the tool. The handler object will be heavily
        # consulted by a plethora of objects during request handling. This is
        # why it is convenient to define such attributes on it.
        self.tool = self.dbConnection.root.objects.get('tool')
        # Call the base handler's method
        Handler.init(self)

    def finish(self):
        '''Closes p_self.rfile and returns the fact that p_self.clientSocket
           must be closed or not.'''
        self.rfile.close()
        return self.closeSocket

    def getLayout(self):
        '''Try to deduce the current layout from the traversal, if present'''
        traversal = getattr(self, 'traversal', None)
        if traversal: return traversal.context.layout

    def isAjax(self):
        '''Is this handler handling an Ajax request ?'''
        traversal = getattr(self, 'traversal', None)
        if traversal and traversal.context: return traversal.context.ajax

    def isPublished(self, o):
        '''Is object p_o the currently published object ?'''
        referer = self.headers.get('referer')
        if not referer: return
        # Ensure the referer URL path ends with a slash
        path = urllib.parse.urlparse(referer).path
        if not path.endswith('/'): path = '%s/' % path
        return ('/%s/' % o.id) in path

    def manageGuardError(self, resp, traversal, error=None):
        '''A security-related error has occurred: the logged user is not allowed
           to access the desired URL. If the user is anon, redirect him to the
           login page. Else, return an error. p_error may contain the raised
           exception instance.'''
        # Return the response content. In the case of a redirect, None will be
        # returned.
        if self.guard.user.isAnon():
            siteUrl = self.config.server.getUrl(self)
            gotoUrl = urllib.parse.quote('%s%s' % (siteUrl, self.path))
            resp.goto(url='%s/tool/home?goto=%s' % (siteUrl, gotoUrl),
                      message=self.tool.translate('please_authenticate'))
            r = None
        else:
            # Log and return a 403 error: forbidden
            resp.code = HTTPStatus.FORBIDDEN
            r = Error.get(resp, traversal, error=error)
        return r

    def manageDatabaseRequest(self):
        '''Manage an HTTP request whose response must imply using the
           database and return the response content as a string.'''
        # Initialise p_self's base data structures
        self.init()
        # Run a traversal
        self.traversal = traversal = Traversal(handler=self)
        resp = self.resp
        try:
            r = traversal.run()
        except Traversal.Error as err:
            resp.code = HTTPStatus.NOT_FOUND # 404
            r = Error.get(resp, traversal)
        except Guard.Error as err:
            r = self.manageGuardError(resp, traversal, error=err)
        except MessageException as msg:
            resp.code = HTTPStatus.OK
            r = Error.get(resp, traversal, error=msg)
        except EvaluationError as err:
            # It can be an exception being re-raised in the context of a PX
            # evaluation error.
            orig = err.originalError
            if isinstance(orig, Guard.Error):
                r = self.manageGuardError(resp, traversal, error=orig)
            elif isinstance(orig, MessageException):
                resp.code = HTTPStatus.OK
                r = Error.get(resp, traversal, error=orig)
            elif orig.__class__ in self.server.database.ConflictErrors:
                # A database conflict error embedded into an evaluation error.
                # Re-raise it as a first class exception: it will be catched and
                # managed by the calling method.
                raise orig
            else:
                resp.code = HTTPStatus.INTERNAL_SERVER_ERROR # 500
                r = Error.get(resp, traversal)
        return r

    def attemptDatabaseRequest(self):
        '''Tries to m_manageDatabaseRequest. Perform several attempts depending 
           on the occurrence of database conflict errors. Returns the content of
           the page to serve, as a string.'''
        maxAttempts = self.config.database.conflictRetries
        attempts = 0
        database = self.server.database
        while True:
            try:
                r = self.manageDatabaseRequest()
                # Perform a database commit when appropriate
                if self.commit: database.commit(self)
            except database.ConflictErrors as err:
                attempts += 1
                # Abort the transaction that could not be committed
                transaction.abort()
                # Return an error if is was the last attempt
                if attempts == maxAttempts:
                    self.log('site', 'error', TRANS_ERR % str(err))
                    self.resp.code = HTTPStatus.SERVICE_UNAVAILABLE # 503
                    r = Error.get(self.resp, self.traversal)
                    break
                else:
                    self.log('site', 'warning',
                             TRANS_RETRY % (str(err), attempts))
            except Exception as err:
                self.resp.code = HTTPStatus.INTERNAL_SERVER_ERROR # 500
                r = Error.get(self.resp, self.traversal)
                break
            else:
                break
        return r

    def determineType(self):
        '''Determines the type of the request and split, in p_self.parts, the
           request path into parts.'''
        self.parsedPath = parsed = urllib.parse.urlparse(self.path.strip('/'))
        parts = parsed.path.split('/')
        config = self.config
        # Remove any blank part among self.parts
        i = len(parts) - 1
        while i >= 0:
            if not parts[i]: del parts[i]
            i -= 1
        # No path at all means: dynamic content must be served (with a default
        # object and default method or PX applied on it).
        if not parts:
            self.static = False
        # If the first part of the path corresponds to the root static folder,
        # static content must be served.
        elif parts[0] == config.server.static.root:
            self.static = True
            del parts[0]
        # If the unique part corresponds to a file (ie, any name containing a
        # dot), it is static content, too (favicon.ico, robots.txt...)
        elif (len(parts) == 1) and ('.' in parts[0]):
            self.static = True
            # This kind of file must exist in the app
            parts.insert(0, config.model.appName)
        # In any other case, dynamic content must be served
        else:
            self.static = False
        self.parts = parts
        # Must the client socket be closed after the request has been handled ?
        httpVersion = config.server.httpVersion
        self.closeSocket = (httpVersion == 1) or \
                           (self.headers.get('Connection') == 'close')

    def handle(self):
        '''Called by m_run to handle a HTTP request'''
        # Determine the type of the request: static or dynamic (boolean
        # self.static). Static content is any not-database-stored-and-editable
        # image, Javascript, CSS file, etc. Dynamic content is any request
        # reading and/or writing to/from the database.
        self.determineType()
        resp = self.resp
        if self.static:
            # Manage static content
            Static.get(self)
        else:
            # Add myself to the registry of handlers
            Handler.add(self)
            # Initialise a database connection
            database = self.server.database
            self.dbConnection = database.openConnection()
            # Extract parameters and create a Request instance in p_self.req
            self.req = Request.create(self)
            # Compute dynamic content
            r = self.attemptDatabaseRequest()   
            # Build the HTTP response (if not done yet)
            try:
                resp.build(r)
            finally:
                # Close the DB connection
                database.closeConnection(self.dbConnection)
                # Remove myself from the registry (for now)
                Handler.remove()
        # Log this hit and the response code on the site log
        self.log('site', 'info', str(resp.code.value))

    def parseRequestHeaders(self):
        '''Parses the request line and headers. Returns (None, None) on success
           and (errorCode, message) on failure.'''
        # Split the request line
        stripped = self.requestLine.rstrip(Response.eol)
        parts = stripped.decode(Response.headersEncoding).split()
        count = len(parts)
        if count not in (2, 3):
            return HTTPStatus.BAD_REQUEST, RQ_LINE_KO % decoded
        # Manage the request version
        if count == 3:
            # The version of the HTTP protocol requested by the client
            version = parts.pop()
            if not version.startswith('HTTP/'):
                return HTTPStatus.BAD_REQUEST, HTTP_V_KO % version
            self.requestVersion = version
            if version.count('.') != 1:
                return HTTPStatus.BAD_REQUEST, HTTP_V_KO % version
            try:
                floatVersion = float(version[5:])
                # Currently, HTTP 1.0 and 1.1 are supported
                if (floatVersion < 1.0) or (floatVersion > 1.1):
                    code = HTTPStatus.HTTP_VERSION_NOT_SUPPORTED
                    return code, HTTP_V_UNS % version
            except ValueError:
                return HTTPStatus.BAD_REQUEST, HTTP_V_KO % version
        else:
            self.requestVersion = self.config.server.getProtocolString()
        # Manage the HTTP method (GET or POST) and path
        self.httpMethod, self.path = parts
        # Parse HTTP headers
        try:
            self.headers = hc.parse_headers(self.rfile, _class=hc.HTTPMessage)
        except hc.LineTooLong as err:
            code = HTTPStatus.REQUEST_HEADER_FIELDS_TOO_LARGE
            return code, HEAD_TOO_LG % str(err)[:15]
        except hc.HTTPException as err:
            code = HTTPStatus.REQUEST_HEADER_FIELDS_TOO_LARGE
            return code, HEAD_T_MANY % str(err)[:15]
        return None, None

    def tlog(self, message, level=1):
        '''Output p_message if the debug level requires it'''
        return self.server.tlog(message, level=level,
                                clientPort=self.clientPort)

    def run(self):
        '''Handles a complete HTTP request. Returns True if p_self.clientSocket
           must be closed afterwards.'''
        try:
            # Read the request line
            self.tlog(T_READ_SOCK)
            self.requestLine = line = self.rfile.readline(65537)
            length = len(line)
            self.tlog(T_READ_GOT % (line, length))
            # Cancel everything if no data was received
            if not line: return self.finish()
            # The request URI may be too long
            if length > 65536:
                self.resp.buildError(HTTPStatus.REQUEST_URI_TOO_LONG)
                return self.finish()
            # Parse request headers
            errorCode, message = self.parseRequestHeaders()
            if errorCode:
                self.resp.buildError(errorCode, message=message)
                return self.finish()
            # Is the HTTP method supported ?
            method = self.httpMethod
            if method not in HttpHandler.supportedHttpMethods:
                self.resp.buildError(HTTPStatus.NOT_IMPLEMENTED,
                                     message=H_METH_KO % method)
                return self.finish()
            # Handle the request
            self.handle()
        except socket.timeout as e:
            # A read or a write timed out. Discard the client socket.
            self.log('app', 'error', SOCK_TO % str(e))
        finally:
            return self.finish()

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class VirtualHandler(Handler):
    '''Fake handler handling a virtual request: server initialization, scheduler
       job, etc.'''

    # Fake handler attributes
    clientPort = 0
    command = 'GET'
    path = '/'
    requestVersion = 'HTTP/1.1'
    headers = {'User-Agent': 'system'}
    fake = True

    # Entries logged from a virtual handler will be prefixed with this char
    logChar = '>'

    def __init__(self, server):
        '''Tries to define the same, or fake version of, a standard handler's
           attributes.'''
        # The Appy HTTP server instance
        self.server = server
        self.config = server.config
        # Create fake Request and Response objects
        self.req = Request()
        self.resp = Response(self)
        # Call the base handler's method
        Handler.init(self)
        # Add myself to the registry of handlers
        Handler.add(self)
        # By default, a commit will be performed
        self.commit = True
        # The following attributes will be initialised later
        self.dbConnection = None
        self.tool = None
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# ------------------------------------------------------------------------------
import sys, time, os, os.path, smtplib, socket, shutil
from optparse import OptionParser
import ZODB.FileStorage
import ZODB.serialize
from DateTime import DateTime
from StringIO import StringIO
from appy.gen.mail import MailConfig, sendMail
from appy.shared.utils import executeCommand
folderName = os.path.dirname(__file__)

# ------------------------------------------------------------------------------
class BackupError(Exception): pass
ERROR_CODE = 1

# ------------------------------------------------------------------------------
class ZodbBackuper:
    '''This backuper will run every night (after 00.00). Every night excepted
       Sunday, it will perform an incremental backup. Every Sunday, the script
       will pack the ZODB, perform a full backup, and, if successful, remove all
       previous (full and incremental) backups.'''
    fullBackupExts = ('.fs', '.fsz')
    allBackupExts = ('.dat', '.deltafs', '.deltafsz') + fullBackupExts
    toRemoveExts = ('.doc', '.pdf', '.odt')
    rsyncOptions = ['rsync', '-Paz', '--no-o', '--no-g', '--no-l', '--delete',
      '--quiet', '--exclude', 'Data.fs*', '--exclude', 'Z2.*']
    # --no-o and --no-g options are for preventing rsync to perform chown and
    # chgrp commands that produce errors when done via an sshfs mount.

    def __init__(self, storageLocation, backupFolder, options):
        self.storageLocation = storageLocation
        self.backupFolder = backupFolder
        self.options = options
        # Unwrap some options directly on self.
        self.repozo = options.repozo or './repozo.py'
        self.zopectl = options.zopectl or './zopectl'
        self.logFile = file(options.logFile, 'a')
        # Keep a log of the last script execution, so we can send this info by
        # email.
        self.logMem = StringIO() 
        self.emails = options.emails.split(',')
        self.tempFolder = options.tempFolder
        self.logsBackupFolder = options.logsBackupFolder
        self.filesBackupFolder = options.filesBackupFolder
        self.zopeUser = options.zopeUser
        self.keepSeconds = int(options.keepSeconds)

    def log(self, msg, type='info'):
        for logPlace in (self.logFile, self.logMem):
            logPlace.write(msg)
            logPlace.write('\n')

    def executeCommand(self, cmd):
        '''Executes command p_cmd'''
        w = self.log
        w('Executing "%s"...' % cmd)
        out, err = executeCommand(cmd)
        if out: w(out)
        if err: w(err)
        w('Done.')

    def packZodb(self):
        '''Packs the ZODB and keeps one week history'''
        storage = ZODB.FileStorage.FileStorage(self.storageLocation)
        #storage.pack(time.time()-(7*24*60*60), ZODB.serialize.referencesf)
        storage.pack(time.time()-self.keepSeconds, ZODB.serialize.referencesf)
        for fileSuffix in ('', '.index'):
            fileName = self.storageLocation + fileSuffix
            executeCommand(['chown', self.zopeUser, fileName])

    def removeDataFsOld(self):
        '''Removes the file Data.fs.old if it exists.

           In the process of packing the ZODB, an additional file Data.fs.pack
           is created, and renamed to Data.fs once finished. It means that, when
           we pack the ZODB, 3 copies of the DB can be present at the same time:
           Data.fs, Data.fs.old and Data.fs.pack. We prefer to remove the
           Data.fs.old copy to avoid missing disk space if the DB is big.
        '''
        old = self.storageLocation + '.old'
        if os.path.exists(old):
            self.log('Removing %s...' % old)
            os.remove(old)
            self.log('Done.')

    folderCreateError = 'Could not create backup folder. Backup of log ' \
                        'files will not take place. %s'
    def backupLogs(self):
        w = self.log
        if not os.path.exists(self.logsBackupFolder):
            # Try to create the folder when to store backups of the log files
            try:
                w('Try to create backup folder for logs "%s"...' % \
                  self.logsBackupFolder)
                os.mkdir(self.logsBackupFolder)
            except IOError, ioe:
                w(folderCreateError % str(ioe))
            except OSError, oe:
                w(folderCreateError % str(oe))
        if os.path.exists(self.logsBackupFolder):
            # Ok, we can make the backup of the log files.
            # Get the folder where logs lie
            logsFolder = self.options.logsFolder
            d = os.path.dirname
            j = os.path.join
            if not logsFolder:
                logsFolder = j(d(d(self.storageLocation)), 'log')
            if not os.path.isdir(logsFolder):
                w('Cannot backup log files because folder "%s" does not ' \
                  'exist. Try using option "-g".' % logsFolder)
                return
            for logFileName in os.listdir(logsFolder):
                if logFileName.endswith('.log'):
                    backupTime = DateTime().strftime('%Y_%m_%d_%H_%M')
                    parts = os.path.splitext(logFileName)
                    copyFileName = '%s.%s%s' % (parts[0], backupTime, parts[1])
                    absCopyFileName = j(self.logsBackupFolder, copyFileName)
                    absLogFileName = j(logsFolder, logFileName)
                    w('Moving "%s" to "%s"...' % (absLogFileName,
                                                  absCopyFileName))
                    shutil.copyfile(absLogFileName, absCopyFileName)
                    os.remove(absLogFileName)
                    # I do a "copy" + a "remove" instead of a "rename" because
                    # a "rename" fails if the source and dest files are on
                    # different physical devices.

    def getDate(self, dateString):
        '''Returns a DateTime instance from p_dateString, which has the form
           YYYY-MM-DD-HH-MM-SS.'''
        return DateTime('%s/%s/%s %s:%s:%s' % tuple(dateString.split('-')))

    def removeOldBackups(self):
        '''This method removes all files (full & incremental backups) that are
           older than the last full backup.'''
        w = self.log
        # Determine date of the oldest full backup
        oldestFullBackupDate = eighties = DateTime('1980/01/01')
        for backupFile in os.listdir(self.backupFolder):
            fileDate, ext = os.path.splitext(backupFile)
            if ext in self.fullBackupExts:
                # I have found a full backup
                fileDate = self.getDate(fileDate)
                if fileDate > oldestFullBackupDate:
                    oldestFullBackupDate = fileDate
        # Remove all backup files older that oldestFullBackupDate
        if oldestFullBackupDate != eighties:
            w('Last full backup date: %s' % str(oldestFullBackupDate))
        for backupFile in os.listdir(self.backupFolder):
            fileDate, ext = os.path.splitext(backupFile)
            if (ext in self.allBackupExts) and \
               (self.getDate(fileDate) < oldestFullBackupDate):
                fullFileName = '%s/%s' % (self.backupFolder, backupFile)
                w('Removing old backup file %s...' % fullFileName)
                os.remove(fullFileName)

    def sendEmails(self):
        '''Send content of self.logMem to self.emails'''
        # Extract server, port, login and password
        login = password = None
        info = self.options.smtpServer.split(':', 3)
        if len(info) == 1: # We simply have the server
            server = info[0]
            port = 25
        elif len(info) == 2: # We have server and port
            server, port = info
            port = int(port)
        else: # We also have login and password
            server, port, login, password = info
        # Secure (TLS) or not ?
        if server.endswith('*'):
            secure = True
            server = server[:-1]
        else:
            secure = False
        # Create a MailConfig instance
        cfg = MailConfig(fromEmail=self.options.fromAddress, server=server,
                       port=port, secure=secure, login=login, password=password)
        # Send the mail
        subject = 'Appy nightlife %s > %s' % \
                  (self.storageLocation, self.backupFolder)
        sendMail(cfg, self.emails, subject, self.logMem.getvalue(),log=self.log)

    def removeTempFiles(self):
        '''LibreOffice (LO) produces temp files that a Appy/Zope process tries
           do delete once he has generated it and returned it to the browser.
           But in some cases it is not possible: Zope runs with a given user
           and LO runs with root and produces files that can't be deleted by the
           user running Zope. This method removes the temp files that could not
           have been removed by Zope.'''
        w = self.log
        w('Removing temp files in "%s"...' % self.tempFolder)
        pdfCount = docCount = odtCount = 0
        for fileName in os.listdir(self.tempFolder):
            ext = os.path.splitext(fileName)[1]
            if ext in self.toRemoveExts:
                exec '%sCount += 1' % ext[1:]
                fullFileName = os.path.join(self.tempFolder, fileName)
                #w('Removing "%s"...' % fullFileName)
                try:
                    os.remove(fullFileName)
                except OSError, oe:
                    w('Could not remove "%s" (%s).' % (fullFileName, str(oe)))
        w('%d .pdf, %d .doc and %d .odt file(s) removed.' % \
          (pdfCount, docCount, odtCount))

    def run(self):
        w = self.log
        startTime = time.time()
        mode = self.options.mode
        w('\n****** Backup launched at %s (mode: %s) ******' % \
          (str(time.asctime()), mode))
        # Shutdown the Zope instance
        w('> Shutting down Zope instance...')
        self.executeCommand([self.zopectl, 'stop'])
        # Check if we are on the "full backup day"
        dayFull = self.options.dayFullBackup
        if (dayFull == 'All') or time.asctime().startswith(dayFull):
            # If mode 'zodb', let's pack the ZODB first. Afterwards it will
            # trigger a full backup.
            if mode == 'zodb':
                # As a preamble to packing the ZODB, remove Data.fs.old if
                # present.
                self.removeDataFsOld()
                w('> Day is "%s", packing the ZODB...' % dayFull)
                self.packZodb()
                w('Pack done.')
            elif mode == 'copy':
                dest = os.path.join(self.backupFolder, 'Data.fs.new')
                w('> Day is "%s", copying %s to %s...' % \
                  (dayFull, self.storageLocation, dest))
                # Perform a copy of Data.fs to the backup folder
                shutil.copyfile(self.storageLocation, dest)
                # The copy has succeeded. Remove the previous copy and rename
                # this one.
                oldDest = os.path.join(self.backupFolder, 'Data.fs')
                w('> Copy successful. Renaming %s to %s...' % (dest, oldDest))
                if os.path.exists(oldDest):
                    os.remove(oldDest)
                    w('> (Old existing backup %s was removed).' % oldDest)
                os.rename(dest, oldDest)
                w('Done.')
            # Make a backup of the log files
            w('> Make a backup of log files...')
            self.backupLogs()
            w('Log files copied.')
        else:
            if mode == 'copy':
                w('Copy mode: nothing to copy: day is not %s.' % dayFull)
        # Do the backup with repozo
        if mode == 'zodb':
            w('> Performing backup...')
            self.executeCommand([self.options.python, self.repozo, '-BvzQ',
                           '-r', self.backupFolder, '-f', self.storageLocation])
            # R-sync the DB-controlled file system
            if self.filesBackupFolder:
                filesFolder = os.path.dirname(self.storageLocation)
                w('> Rsyncing DB-controlled file system @%s...' % filesFolder)
                command = self.rsyncOptions + [filesFolder + '/',
                                               self.filesBackupFolder]
                self.executeCommand(command)
                w('Done.')
            # Remove previous full backups
            self.removeOldBackups()
        # If a command is specified, run Zope to execute this command
        if self.options.command:
            w('> Executing command "%s"...' % self.options.command)
            jobScript = '%s/job.py' % folderName
            self.executeCommand([self.zopectl, 'run', jobScript,
                                 self.options.command])
        self.removeTempFiles()
        # Start the instance again, in normal mode
        w('> Restarting Zope instance...')
        self.executeCommand([self.zopectl, 'start'])
        stopTime = time.time()
        w('Done in %d minute(s).' % ((stopTime-startTime)/60))
        if self.emails:
            self.sendEmails()
        self.logFile.close()
        print(self.logMem.getvalue())
        self.logMem.close()

# ------------------------------------------------------------------------------
class ZodbBackupScript:
    '''usage: python backup.py storageLocation backupFolder [options]
       storageLocation is the path to a ZODB database (file storage) (ie
                       /opt/ZopeInstance/var/Data.fs);
       backupFolder is a folder exclusively dedicated for storing backups
                      of the mentioned storage (ie /data/zodbbackups).'''

    weekDays = ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')
    def checkArgs(self, options, args):
        '''Check that the scripts arguments are correct.'''
        # Do I have the correct number of args?
        if len(args) != 2:
            raise BackupError('Wrong number of arguments.')
        # Check storageLocation
        if not os.path.exists(args[0]) or not os.path.isfile(args[0]):
            raise BackupError('"%s" does not exist or is not a file.' % args[0])
        # Check backupFolder
        if not os.path.isdir(args[1]):
            raise BackupError('"%s" does not exist or is not a folder.'%args[1])
        # Check logs folder
        if options.logsFolder and not os.path.isdir(options.logsFolder):
            raise BackupError('"%s" is not a folder.' % options.logsFolder)
        # Check files folder
        filesFolder = options.filesBackupFolder
        if filesFolder and not os.path.isdir(filesFolder):
            raise BackupError('"%s" is not a folder.' % filesFolder)
        # Try to create a file in this folder to check if we have write
        # access in it.
        fileName = '%s/%s.tmp' % (args[1], str(time.time()))
        try:
            f = file(fileName, 'w')
            f.write('Hello.')
            f.close()
            os.remove(fileName)
        except OSError, oe:
            raise BackupError('I do not have the right to write in ' \
                              'folder "%s".' % args[1])
        # Check temp folder
        if not os.path.isdir(options.tempFolder):
            raise BackupError('Temp folder "%s" does not exist or is not ' \
                              'a folder.' % options.tempFolder)
        # Check day of week
        dfb = options.dayFullBackup
        if (dfb != 'All') and (dfb not in self.weekDays):
            raise BackupError(
                "Day of week must be one of %s or 'All'." % str(self.weekDays))
        # Check command format
        if options.command:
            parts = options.command.split(':')
            if len(parts) not in (4,5):
                raise BackupError('Command format must be ' \
                    '<ZopeAdmin><PloneInstancePath>:<ApplicationName>:' \
                    '<ToolMethodName>[:<args>]')

    def run(self):
        optParser = OptionParser(usage=ZodbBackupScript.__doc__)
        optParser.add_option("-p", "--python", dest="python",
            help="The path to the Python interpreter running Zope",
            default='python2.4',metavar="PYTHON",type='string')
        optParser.add_option("-r", "--repozo", dest="repozo",
            help="The path to repozo.py", default='', metavar="REPOZO",
            type='string')
        optParser.add_option("-z", "--zopectl", dest="zopectl",
            help="The path to Zope instance's zopectl script", default='',
            metavar="ZOPECTL", type='string')
        optParser.add_option("-l", "--logfile", dest="logFile",
            help="Log file where this script will append output (defaults to " \
                 "./backup.log)", default='./backup.log', metavar="LOGFILE",
            type='string')
        optParser.add_option("-d", "--day-full-backup", dest="dayFullBackup",
            help="Day of the week where the full backup must be performed " \
                 "(defaults to 'Sun'). Must be one of %s or 'All' if you " \
                 "want a full backup everyday." % str(self.weekDays),
            default='Sun', metavar="DAYFULLBACKUP", type='string')
        optParser.add_option("-e", "--emails", dest="emails",
            help="Comma-separated list of emails that will receive the log " \
                 "of this script.", default='', metavar="EMAILS", type='string')
        optParser.add_option("-f", "--from-address", dest="fromAddress",
            help="From address for the sent mails", default='',
            metavar="FROMADDRESS", type='string')
        optParser.add_option("-s", "--smtp-server", dest="smtpServer",
            help="SMTP server and port (ie: localhost:25) for sending mails. " \
                 "You can also embed username and password if the SMTP " \
                 "server requires authentication, ie localhost:25:myLogin:" \
                 "myPassword. If you want a secure connection to the SMTP " \
                 "server (TLS), use a star, ie secure.smtp.server*:25.",
            default='localhost:25', metavar="SMTPSERVER", type='string')
        optParser.add_option("-t", "--tempFolder", dest="tempFolder",
            help="Folder used by LibreOffice for producing temp files. " \
                 "Defaults to /tmp.", default='/tmp', metavar="TEMP",
            type='string')
        optParser.add_option("-g", "--logsFolder",dest="logsFolder",
            help="Folder where Zope log files are (typically: event.log and " \
                 "Z2.log). If no folder is provided, we will consider to " \
                 "work on a standard Zope instance and decide that the log " \
                 "folder is, from 'storageLocation', located at ../log",
            metavar="LOGSFOLDER", type='string')
        optParser.add_option("-b", "--logsBackupFolder",dest="logsBackupFolder",
            help="Folder where backups of log files (event.log and Z2.log) " \
            "will be stored.", default='./logsbackup',
            metavar="LOGSBACKUPFOLDER", type='string')
        optParser.add_option("-F", "--filesBackupFolder",
            dest="filesBackupFolder",
            help="Folder where backups of the DB-controlle file hierarchy " \
                 "will be stored.", default='',
                 metavar="FILESBACKUPFOLDER", type='string')

        optParser.add_option("-u", "--user", dest="zopeUser",
            help="User and group that must own Data.fs. Defaults to " \
                 "zope:www-data. If this script is launched by root, for " \
                 "example, when packing the ZODB this script may produce a " \
                 "new Data.fs that the user running Zope may not be able to " \
                 "read anymore. After packing, this script makes a 'chmod' " \
                 "on Data.fs.", default='zope:www-data', metavar="USER",
            type='string')
        optParser.add_option("-k", "--keep-seconds", dest="keepSeconds",
            help="Number of seconds to leave in the ZODB history when the " \
                 "ZODB is packed.", default='86400', metavar="KEEPSECONDS",
            type='string')
        optParser.add_option("-m", "--mode", dest="mode", help="Default mode, "\
            "'zodb', uses repozo for performing backups. Mode 'copy' simply " \
            "performs a copy of the database to the specified backup folder.",
            default='zodb', metavar="MODE", type='string')
        optParser.add_option("-c", "--command", dest="command",
            help="Command to execute while Zope is running. It must have the " \
            "following format: <ZopeAdmin>:<PloneInstancePath>:" \
            "<ApplicationName>:<ToolMethodName>[:<args>]. <ZopeAdmin> is the " \
            "user name of the Zope administrator; <PloneInstancePath> is the " \
            "path, within Zope, to the Plone Site object (if not at the " \
            "root of the Zope hierarchy, use '/' as folder separator); " \
            "<ApplicationName> is the name of the Appy application; " \
            "<ToolMethodName> is the name of the method to call on the tool " \
            "in this Appy application; (optional) <args> are the arguments " \
            "to give to this method (only strings are supported). Several " \
            "arguments must be separated by '*'.",  default='',
            metavar="COMMAND", type='string')
        (options, args) = optParser.parse_args()
        try:
            self.checkArgs(options, args)
            backuper = ZodbBackuper(args[0], args[1], options)
            backuper.run()
        except BackupError, be:
            sys.stderr.write(str(be) + '\nRun the script without any ' \
                                       'argument for getting help.\n')
            sys.exit(ERROR_CODE)

# ------------------------------------------------------------------------------
if __name__ == '__main__':
    ZodbBackupScript().run()
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.model.fields import Field

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Color(Field):
    '''Field representing a color'''

    view = cell = Px('''
     <input type="color" disabled=":True" value=":value or '#000000'"
            name=":name" id=":name"/>''')

    edit = search = Px('''
     <input type="color" id=":name" name=":name" size=":field.width"
            value=":field.getInputValue(inRequest, requestValue, value)"/>
     <script if="hostLayout">:'prepareForAjaxSave(%s,%s,%s,%s)' % \
      (q(name), q(o.iid), q(o.url), q(hostLayout))</script>''')

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, indexed=False, mustIndex=True, indexValue=None,
      emptyIndexValue='-', searchable=False, filterField=None,
      readPermission='read', writePermission='write', width=5, height=None,
      maxChars=13, colspan=1, master=None, masterValue=None, focus=False,
      historized=False, mapping=None, generateLabel=None, label=None,
      sdefault=None, scolspan=1, swidth=None, sheight=None, persist=True,
      inlineEdit=False, view=None, cell=None, edit=None, xml=None,
      translations=None):
        # Call the base constructor
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          emptyIndexValue, searchable, filterField, readPermission,
          writePermission, width, height, maxChars, colspan, master,
          masterValue, focus, historized, mapping, generateLabel, label,
          sdefault, scolspan, swidth, sheight, persist, inlineEdit, view, cell,
          edit, xml, translations)
        # Define the corresponding Python type for values stored for this type
        self.pythonType = str
        # The "Color" HTML field requires a default value
        if not default:
            self.default = '#000000' # Black
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Global elements to include in HTML pages'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.ui.iframe import Iframe

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Globals:
    '''Global elements to inject in most or all pages'''

    # Translated messages computed in Javascript variables in most pages
    variables = ('no_elem_selected', 'action_confirm', 'save_confirm',
                 'warn_leave_form', 'workflow_comment')

    @classmethod
    def getVariables(class_, tool):
        '''Returns Javascript variables storing translated texts used in forms
           and popups.'''
        r = ['var wrongTextInput="%s none";' % tool.config.ui.wrongTextColor]
        for label in class_.variables:
            r.append('var %s="%s";' % (label, tool.translate(label)))
        return '<script>%s</script>' % '\n'.join(r)

    # Popups must be present in every page
    popups = '''%s
     <!-- Popup for confirming an action -->
     <div id="confirmActionPopup" class="popup">
      <img class="iconM popupI" src="%s"/>
      <form id="confirmActionForm" method="post">
       <div align="center">
        <p id="appyConfirmText"></p>
        <input type="hidden" name="actionType"/>
        <input type="hidden" name="action"/>
        <input type="hidden" name="visible"/>
        <div id="commentArea" align="%s">
         <div id="appyCommentLabel"></div>
         <textarea name="popupComment" id="popupComment"
                   cols="30" rows="3"></textarea>
        </div>
        <div class="topSpace">
         <input type="button" onclick="doConfirm()" value="%s"/>
         <input type="button" value="%s"
                onclick="closePopup('confirmActionPopup', 'comment')"/>
        </div>
       </div>
      </form>
     </div>
     <!-- Popup for uploading a file in a pod field -->
     <div id="uploadPopup" class="popup" align="center">
      <form id="uploadForm" name="uploadForm" enctype="multipart/form-data"
            method="post">
       <input type="hidden" name="objectId"/>
       <input type="hidden" name="fieldName"/>
       <input type="hidden" name="template"/>
       <input type="hidden" name="podFormat"/>
       <input type="hidden" name="action" value="upload"/>
       <input type="file" name="uploadedFile"/><br/><br/>
       <input type="submit" value="%s"/>
       <input type="button" onclick="closePopup('uploadPopup')" value="%s"/>
      </form>
     </div>
     <!-- Popup for displaying an error message -->
     <div id="alertPopup" class="popup">
      <img class="iconM popupI" src="%s"/>
      <div align="center" id="appyAlertText"></div>
      <div align="center" class="topSpace">
       <input type="button" onclick="closePopup('alertPopup')" value="%s"/>
      </div>
     </div>%s'''

    @classmethod
    def getPopups(class_, tool, url, _, dleft, dright, popup):
        '''Returns the popups to include in every page'''
        # The "iframe" popup must not be included if we are already in a popup
        if popup:
            iframe = ''
        else:
            iframe = Iframe.view % (url('close.svg'), url('arrows.svg'))
        # Define variables, per popup
        vars = (
         # Global Javascript variables
         class_.getVariables(tool),
         # confirmActionPopup
         url('arrows.svg'), dleft, _('yes'), _('no'),
         # uploadPopup
         _('object_save'), _('object_cancel'),
         # alertPopup
         url('warning.svg'), _('appy_ok'),
         # iframePopup
         iframe
        )
        return class_.popups % vars

    # Forms must be present on some pages, like view, edit and search.
    # Global Javascript message are also defined here.
    forms = '''
     <!-- Global form for editing entries within an object's history -->
     <form id="eventForm" method="post">
      <input type="hidden" name="objectId"/>
      <input type="hidden" name="eventTime"/>
      <input type="hidden" name="comment"/>
     </form>
     <!-- Global form for generating/freezing a document from a pod template -->
     <form id="podForm" name="podForm" method="post">
      <input type="hidden" name="template"/>
      <input type="hidden" name="podFormat"/>
      <input type="hidden" name="queryData"/>
      <input type="hidden" name="criteria"/>
      <input type="hidden" name="customParams"/>
      <input type="hidden" name="showSubTitles" value="True"/>
      <input type="hidden" name="checkedIds"/>
      <input type="hidden" name="checkedSem"/>
      <input type="hidden" name="freezeAction"/>
      <input type="hidden" name="mailing"/>
      <input type="hidden" name="mailText"/>
     </form>'''

    @classmethod
    def getForms(class_, tool):
        '''Returns the forms to include in most pages, as well as translated
           Javascript messages.'''
        return class_.forms

    @classmethod
    def getScripts(class_, tool, q, layout):
        '''Get the scripts that must be run on most pages'''
        return '<script>initSlaves(%s,%s)</script>' % (q(tool.url), q(layout))
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import re, utils
from appy.shared.utils import formatNumber

htmlColorNames = {
  'aliceblue': '#f0f8ff', 'antiquewhite': '#faebd7', 'aqua': '#00ffff',
  'aquamarine': '#7fffd4','azure': '#f0ffff', 'beige': '#f5f5dc',
  'bisque': '#ffe4c4', 'black': '#000000', 'blanchedalmond': '#ffebcd',
  'blue': '#0000ff', 'blueviolet': '#8a2be2', 'brown': '#a52a2a',
  'burlywood': '#deb887', 'cadetblue': '#5f9ea0', 'chartreuse': '#7fff00',
  'chocolate': '#d2691e', 'coral': '#ff7f50', 'cornflowerblue': '#6495ed',
  'cornsilk': '#fff8dc', 'crimson': '#dc143c', 'cyan': '#00ffff',
  'darkblue': '#00008b', 'darkcyan': '#008b8b', 'darkgoldenrod': '#b8860b',
  'darkgray': '#a9a9a9', 'darkgrey': '#a9a9a9', 'darkgreen': '#006400',
  'darkkhaki': '#bdb76b', 'darkmagenta': '#8b008b', 'darkolivegreen': '#556b2f',
  'darkorange': '#ff8c00', 'darkorchid': '#9932cc', 'darkred': '#8b0000',
  'darksalmon': '#e9967a', 'darkseagreen': '#8fbc8f', 'darkslateblue':'#483d8b',
  'darkslategray': '#2f4f4f', 'darkslategrey': '#2f4f4f',
  'darkturquoise': '#00ced1', 'darkviolet': '#9400d3', 'deeppink': '#ff1493',
  'deepskyblue': '#00bfff', 'dimgray': '#696969', 'dimgrey': '#696969',
  'dodgerblue': '#1e90ff', 'firebrick': '#b22222', 'floralwhite': '#fffaf0',
  'forestgreen': '#228b22', 'fuchsia': '#ff00ff', 'gainsboro': '#dcdcdc',
  'ghostwhite': '#f8f8ff', 'gold': '#ffd700', 'goldenrod': '#daa520',
  'gray': '#808080', 'grey': '#808080', 'green': '#008000',
  'greenyellow': '#adff2f', 'honeydew': '#f0fff0', 'hotpink': '#ff69b4',
  'indianred ': '#cd5c5c', 'indigo ': '#4b0082', 'ivory': '#fffff0',
  'khaki': '#f0e68c', 'lavender': '#e6e6fa', 'lavenderblush': '#fff0f5',
  'lawngreen': '#7cfc00', 'lemonchiffon': '#fffacd', 'lightblue': '#add8e6',
  'lightcoral': '#f08080', 'lightcyan': '#e0ffff',
  'lightgoldenrodyellow': '#fafad2', 'lightgray': '#d3d3d3',
  'lightgrey': '#d3d3d3', 'lightgreen': '#90ee90', 'lightpink': '#ffb6c1',
  'lightsalmon': '#ffa07a', 'lightseagreen': '#20b2aa',
  'lightskyblue': '#87cefa', 'lightslategray': '#778899',
  'lightslategrey': '#778899', 'lightsteelblue': '#b0c4de',
  'lightyellow': '#ffffe0', 'lime': '#00ff00', 'limegreen': '#32cd32',
  'linen': '#faf0e6', 'magenta': '#ff00ff', 'maroon': '#800000',
  'mediumaquamarine': '#66cdaa', 'mediumblue': '#0000cd',
  'mediumorchid': '#ba55d3', 'mediumpurple': '#9370db',
  'mediumseagreen': '#3cb371', 'mediumslateblue': '#7b68ee',
  'mediumspringgreen': '#00fa9a', 'mediumturquoise': '#48d1cc',
  'mediumvioletred': '#c71585', 'midnightblue': '#191970',
  'mintcream': '#f5fffa', 'mistyrose': '#ffe4e1', 'moccasin': '#ffe4b5',
  'navajowhite': '#ffdead', 'navy': '#000080', 'oldlace': '#fdf5e6',
  'olive': '#808000', 'olivedrab': '#6b8e23', 'orange': '#ffa500',
  'orangered': '#ff4500', 'orchid': '#da70d6', 'palegoldenrod': '#eee8aa',
  'palegreen': '#98fb98', 'paleturquoise': '#afeeee','palevioletred': '#db7093',
  'papayawhip': '#ffefd5', 'peachpuff': '#ffdab9', 'peru': '#cd853f',
  'pink': '#ffc0cb', 'plum': '#dda0dd', 'powderblue': '#b0e0e6',
  'purple': '#800080', 'rebeccapurple': '#663399', 'red': '#ff0000',
  'rosybrown': '#bc8f8f', 'royalblue': '#4169e1', 'saddlebrown': '#8b4513',
  'salmon': '#fa8072', 'sandybrown': '#f4a460', 'seagreen': '#2e8b57',
  'seashell': '#fff5ee', 'sienna': '#a0522d', 'silver': '#c0c0c0',
  'skyblue': '#87ceeb', 'slateblue': '#6a5acd', 'slategray': '#708090',
  'slategrey': '#708090', 'snow': '#fffafa', 'springgreen': '#00ff7f',
  'steelblue': '#4682b4', 'tan': '#d2b48c', 'teal': '#008080',
  'thistle': '#d8bfd8', 'tomato': '#ff6347', 'turquoise': '#40e0d0',
  'violet': '#ee82ee', 'wheat': '#f5deb3', 'white': '#ffffff',
  'whitesmoke': '#f5f5f5', 'yellow': '#ffff00', 'yellowgreen': '#9acd32'}

# ------------------------------------------------------------------------------
# Default ratio for converting pixels to cm
px2cm = 44.173513561

# ------------------------------------------------------------------------------
ADD_CLASS_ERROR = 'Use method addClass to add CSS classes to this instance'
MERGE_VALUE_ERROR = 'Non-string CSS values cannot be merged or unmixed.'

# ------------------------------------------------------------------------------
def parseStyleAttribute(value, asDict=False):
    '''Returns a list of CSS (name, value) pairs (or a dict if p_asDict is
       True), parsed from p_value, which holds the content of a HTML "style"
       tag.'''
    if asDict: res = {}
    else:      res = []
    for attr in value.split(';'):
        if not attr.strip(): continue
        name, value = attr.split(':', 1)
        if asDict: res[name.strip()] = value.strip()
        else:      res.append( (name.strip(), value.strip()) )
    return res

# ------------------------------------------------------------------------------
class CssValue:
    '''Represents a CSS value'''
    # CSS properties having a unit, with their default unit
    unitProperties = {'width': 'px', 'height': 'px', 'margin-left': 'px',
      'margin-right': 'px', 'margin-top': 'px', 'margin-bottom': 'px',
      'text-indent': 'px', 'font-size': None, 'border-spacing': 'px',
      'line-height': ''} # line-height can be "normal" or a value with a unit
    # CSS properties defining colors
    colorProperties = ('color', 'background-color')
    # Regular expressions for parsing (parts of) CSS values
    valueRex = re.compile('(-?\d*(?:\.\d+)?)(%|px|cm|pt|em|ex)?')
    rgbVal = '\s*([\d.]+)\s*'
    rgbRex = re.compile('rgba?\(%s,%s,%s(?:,%s)?\)' % ((rgbVal,)*4))

    def __init__(self, name, value):
        # p_value can be another CssValue instance
        if isinstance(value, CssValue):
            self.value = value.value
            self.unit = value.unit
            return
        # If we are here, p_value is a string
        self.unit = None
        value = value.strip().lower()
        if name in CssValue.unitProperties:
            val, unit = CssValue.valueRex.match(value).groups()
            if not val:
                # Could not parse it. Maybe a enum value or a not-supported-yet
                # unit.
                self.value = value
            else:
                self.value = float(val)
            self.unit = unit or CssValue.unitProperties[name]
        elif name in CssValue.colorProperties:
            if value.startswith('#'):
                self.value = value # Hexadecimal, keep it as is
            elif value.startswith('rgb'):
                # Convert the RGB value to hexadecimal
                self.value = self.rgb2hex(value)
            else:
                # Probably a color name. Convert it to hexadecimal.
                self.value = htmlColorNames.get(value, value)
        else:
            self.value = value

    def rgb2hex(self, value):
        '''Converts a color expressed in RGB to hexadecimal'''
        match = CssValue.rgbRex.match(value)
        # If we couldn't parse the value, left it untouched
        if not match: return value
        res = '#'
        i = -1
        for val in match.groups():
            i += 1
            # Ignore the optional alpha channel
            if i > 2: continue
            # Convert the integer value to hexadecimal
            hexa = hex(int(val))[2:].upper().zfill(2) # Remove prefix "0x"
            res += hexa
        return res

    def __nonzero__(self):
        # This value is considered empty when self.value is an empty string
        return not isinstance(self.value, basestring) or bool(self.value)

    def merge(self, value):
        '''Add a new p_value among self.value, that can be a space-separated
           list of values.'''
        if not isinstance(self.value, basestring):
            raise Exception(MERGE_VALUE_ERROR)
        self.value = utils.sadd(self.value, value)

    def unmix(self, value):
        '''Removes p_value from self.value, that can be a space-separated list
           of values.'''
        if not isinstance(self.value, basestring):
            raise Exception(MERGE_VALUE_ERROR)
        self.value = utils.sremove(self.value, value)

    def isMultiple(self):
        '''Returns True if self.value contains several sub-values'''
        val = self.value
        return isinstance(val, basestring) and (' ' in val)

    def __str__(self):
        res = str(self.value)
        if self.unit: res += self.unit
        return res

    def __repr__(self): return self.__str__()

    def cm(self, formatted=True, value=None, ratio=px2cm):
        '''Returns p_self.value (or the given p_value) expressed in cm'''
        unit = self.unit
        if not unit: return
        if value is None:
            r = self.value
        else:
            r = value
        if unit == 'px': # Convert pixels to cm
            r = r / ratio
        elif unit in ('em', 'ex'):
            # For "ex", the ratio must be adapted
            if unit == 'ex': ratio = ratio * 1.6
            r = (r / ratio) * 10
        elif unit == 'pt':
            # Use a specific ratio
            r *= 0.0352778
        elif unit != 'cm':
            r = None
        if formatted and (r is not None):
            r = formatNumber(r, sep='.', precision=3)
        return r

# ------------------------------------------------------------------------------
class CssStyles:
    '''This class represents a set of styles collected from:
       * an HTML "style" attribute;
       * other attributes like "width";
       * an HTML tag itself (ie, <b>, <i>...).
       Moreover, it can refer to externally defined CSS classes (mentioned in a
       "class" attribute) in its attribute "classes".
    '''
    # The correspondance between XHTML attributes and CSS properties. Within
    # CSS property names, dashes have been removed because they are used as
    # names for Python instance attributes.
    xhtml2css = {'width': 'width', 'height': 'height', 'align': 'text-align',
                 'cellspacing': 'border-spacing', 'border': 'border'}
    # CSS properties whose values can be combined
    combinable = ('textdecoration',)
    # CSS combined properties (= that can be split into individual properties)
    combined = ('margin',)
    # Directions in use in combined CSS properties (padding, margin, etc),
    # defined in the standard order.
    directions = ('top', 'right', 'bottom', 'left')
    # Values, on combined attributes, that prevent splitting
    unsplittable = ('auto',)
    # Values that must be ignored
    ignore = {'*': {'auto':None, 'initial':None, 'inherit':None}}
    # Values specified in ignore['*'] must not be ignored for these attributes
    ignoreExcept = {'auto': {'table-layout':None}}
    # The '!important' CSS rule
    importantRule = '!important'

    def __init__(self, elem=None, attrs=None, other=None, **kwargs):
        '''Analyses styles as found in p_attrs and/or deduced from XHTML element
           p_elem, and sets, for every found style, an attribute on self. This
           constructor can also be used to build a CssStyles instance from
           another CssStyles instance, given in p_other. A CssStyles instance
           can also be initialised from p_kwargs representing CSS attributes.
           CSS styles defined in p_kwargs override those in p_other, that, in
           turn,override any existing value. Within p_kwargs, CSS attribute
           names must not contain any dash (it would produce illegal Python
           code).'''
        # The content of a potential "class" attribute in p_attrs
        self.classes = ''
        # Parse first the "style" attribute if present
        if attrs and attrs.has_key('style'):
            value = attrs['style']
            if value:
                styles = parseStyleAttribute(attrs['style'], asDict=True)
                for name, value in styles.iteritems():
                    self.add(name, value)
        # Parse obsolete XHTML style-related attributes if present. But they
        # will not override corresponding attributes from the "styles"
        # attributes if found.
        if attrs:
            for name, value in attrs.items():
                if value and (name in self.xhtml2css):
                    self.add(self.xhtml2css[name], value, override=False)
        # If a "class" attribute is defined, store, in attribute "classes", the
        # "external" class(es) defined in it.
        if attrs and attrs.has_key('class'):
            css = attrs['class']
            if css: self.addClass(css)
        # Add the style(s) corresponding to p_elem when relevant, excepted if
        # the corresponding CssValue is already defined.
        if elem and (elem in self.tag2css):
            self.merge(self.tag2css[elem], override=False)
        # Get the styles from p_other and/or p_kwargs. Those styles will
        # override any exsting one.
        if other: self.merge(other)
        for k, v in kwargs.iteritems(): self.add(k, v)
        # Split combined CSS properties
        self.split()

    def __repr__(self):
        res = '<CSS'
        for name, value in self.__dict__.iteritems():
            if name == 'classes':
                if value: res += ' %s::%s' % (name, value)
            else:
                res += ' %s:%s' % (name, value)
        return res + '>'

    def __nonzero__(self):
        # Count all attributes but "classes" that is always present
        return len(self.__dict__) > 1 or bool(self.classes)

    def get(self, name=None):
        '''Gets the value (as a CssValue instance) corresponding to the CSS
           attribute p_name. If p_name is ommitted, it returns all CSS
           attributes as a dict. This method allows to use the original CSS
           attribute names (possibly containing dashes). If this CssStyles
           instance does not store such an attribute, no error is raised:
           None is returned.'''
        if not name: return self.__dict__
        if '-' in name: name = name.replace('-', '')
        return getattr(self, name, None)

    def addClass(self, value, append=True):
        '''Add CSS class(es) p_value (as a string) to self.classes. p_value can
           hold several CSS space-separated classes.'''
        # If p_append is True, CSS classes in p_value are appended to
        # p_self.classes. Else, they are inserted at the start of
        # p_self.classes.
        self.classes = utils.sadd(self.classes, value, append=append)

    def deleteClass(self, value):
        '''Remove CSS class(es) in p_value from self.classes'''
        self.classes = utils.sremove(self.classes, value)

    def getClass(self, last=False):
        '''Returns the name of the CSS class(es) defined in p_self.classes, or
           None if no class is defined. If p_last is True, and several classes
           are defined, only the last one is returned.'''
        # Is a CSS class defined ?
        classes = self.classes
        if not classes: return
        # Yes. Keep only the last one if requested.
        if last and (' ' in classes):
            classes = classes.split()[-1]
        return classes

    def hasClass(self, css):
        '''Is CSS class p_css defined ?'''
        classes = self.classes
        if not classes: return
        if ' ' in classes:
            r = css in classes.split()
        else:
            r = css == classes
        return r

    def add(self, name, value, override=True):
        '''Adds a CSS attribute to this instance as a CssValue instance, and
           also return it. p_name can be the original name of the CSS attribute
           (possibly containing dashes).'''
        if name == 'classes': raise Exception(ADD_CLASS_ERROR)
        # Ignore potential "!important" suffix
        if isinstance(value, basestring) and value.endswith(self.importantRule):
            value = value[:-len(self.importantRule)].strip()
        # Do not add anything if the value must be ignored
        ignore = self.ignore
        if (value in ignore['*']) or \
           ((name in ignore) and (value in ignore[name])):
               # Ignore it, excepted if there is a match in "ignoreExcept"
               if (value not in self.ignoreExcept) or \
                  (name not in self.ignoreExcept[value]):
                   return
        attrName = name.replace('-', '')
        if not override and hasattr(self, attrName): return
        # Define the new value or combine it with the existing value when
        # relevant.
        res = getattr(self, attrName, None)
        if res and (attrName in CssStyles.combinable):
            res.merge(value.value)
        else:
            res = CssValue(name, value)
            setattr(self, attrName, res)
        return res

    def delete(self, name, value=None):
        '''Removes CSS attribute p_name from self, or only p_value for
           combinable CSS properties.'''
        attrName = name.replace('-', '')
        # Nothing to do if the corresponding attribute is not defined
        if not hasattr(self, attrName): return
        # Manage value removal from a combinable CSS property
        existing = getattr(self, attrName)
        if existing and value and (name in CssStyles.combinable):
            existing.unmix(value.value)
            if not existing:
                delattr(self, attrName)
        else:
            delattr(self, attrName)

    def merge(self, other, override=True):
        '''Merges other's styles within self's styles'''
        for name, value in other.get().iteritems():
            if name == 'classes':
                if value:
                    self.addClass(value)
            else:
                self.add(name, value, override=override)

    def unmix(self, other):
        '''Ensures styles defined in p_other are not in p_self'''
        for name, value in other.get().iteritems():
            if name == 'classes':
                self.deleteClass(value)
            else:
                self.delete(name, value)

    def split(self):
        '''Splits combined properties found on p_self into individual
           properties.'''
        # For every combined CSS property found on p_self:
        #  - if it refers to an "unsplittable" value (like "margin:auto"), the
        #    property is removed from p_self and no other property is added;
        #  - if it refers to a "splittable" value, the property is removed from
        #    p_self and the corresponding individual values (one for each
        #    direction as specified in CssStyles.directions) is added to p_self.
        for prop in CssStyles.combined:
            if not hasattr(self, prop): continue
            value = getattr(self, prop)
            if not value or (value.value in CssStyles.unsplittable):
                delattr(self, prop)
                continue
            value = value.value.split()
            length = len(value)
            if length == 1:
                value = value * 4
            elif length == 2:
                value = (value[0], value[1], value[0], value[1])
            elif length == 3:
                value = (value[0], value[1], value[2], value[1])
            elif length != 4:
                # This will be ignored
                delattr(self, prop)
                continue
            # Replace the combined property with individual ones
            i = -1
            for direction in CssStyles.directions:
                i += 1
                self.add('%s-%s' % (prop, direction), value[i])
            delattr(self, prop)

# ------------------------------------------------------------------------------
# For some XHTML tags, we define CssStyle instances containing one or several
# styles that correspond to them.
bold = CssStyles(fontweight='bold')
italic = CssStyles(fontstyle='italic')
strike = CssStyles(textdecoration='line-through')

CssStyles.tag2css = {'b': bold, 'strong': bold, 'i': italic, 'em': italic,
  'strike': strike, 's': strike, 'u': CssStyles(textdecoration='underline'),
  'sup': CssStyles(verticalalign='super'), 'sub': CssStyles(verticalalign='sub')
}
# ------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import re

from appy.px import Px
from appy.xml import Parser
from appy.utils import flipDict
from appy.ui.layout import Layouts
from appy.xml.escape import Escape
from appy.model.fields import Field
from appy.utils import string as sutils
from appy.ui.layout import Layouts, Layout
from appy.database.operators import or_, in_
from appy.database.indexes.text import TextIndex
from appy.model.fields.multilingual import Multilingual

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
INLINE_EDIT_MULTILINGUAL = 'Is is currently not possible to inline-edit ' \
                           'multilingual Text fields.'

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Replacements:
    '''Used by class Text2Html below, this class is used to define replacements
       for characters found in structured text, in order to recreate formatting
       from structured text written with some conventions.'''
    # Default replacement conventions
    default = {
      # Replace double quotes by "guillemets" (angle quotes). The sub-dict has
      # this meaning: if the previous char is a space (or if there is no
      # previous char), the double quote will be replaced with an opening
      # guillemet; else (special key "0"), it will be replaced by a closing
      # guillemet. The guillemet is flanked by a non-breakable space.
      '"': {' ': '« ', '': '« ', 0: ' »'},
      # Conventions for bold, italic and highlighting
      '[': '<b>', ']': '</b>',
      '<': '<i>', '>': '</i>',
      '{': '<span style="background-color: yellow">', '}': '</span>',
      '&': '&amp;' # XML escaping
    }
    # Among "default", some keys represent opening and ending delimiters
    delimiters = {'[': ']', '<': '>', '{': '}'}
    endDelimiters = flipDict(delimiters)
    # Prefix (semi)colons, question and exclamation marks with a non-breakable
    # space, excepted if such a space is already present.
    nbspChars = (':', ';', '!', '?', '%')
    for c in nbspChars:
        default[c] = {' ': c, 0: ' %s' % c}
    whitespace = (' ', ' ')
    # Regular expression for replacing special markers (see "replacementFun"
    # attribute below.
    regex = re.compile('\*(.+?)\*')

    def __init__(self, replacements=None):
        # Apply custom replacements (if any). In order to disable a default
        # replacement, add it in p_replacement with a "None" value.
        if replacements:
            self.replacements = Replacements.default.copy()
            self.replacements.update(replacements)
        else:
            self.replacements = Replacements.default

    def getReplacementFor(self, char, previous):
        '''Return the replacement char for p_char. p_previous is the previous
           char encountered before p_char.'''
        r = self.replacements[char]
        if not isinstance(r, basestring):
            # The replacement depends on "previous"
            if previous in r:
                r = r[previous]
            else:
                r = r[0]
        return r

    def applyOn(self, text, fun=None, toReopen=None):
        '''Apply replacements on p_text and return a tuple
                           (s_result, [s_toReopen])
        '''
        # ----------------------------------------------------------------------
        #  s_result  | Is the resulting unicode string, whose special chars have
        #            | been replaced with their replacements.
        # ----------------------------------------------------------------------
        # [toReopen] | Among special chars to replace, there are opening and
        #            | ending chars, used as delimiters for applying style, ie,
        #            | bold and italic. After having processed p_text, for some
        #            | opening delimiters, we may not have encountered their
        #            | corresponding ending delimiter. If it occurs, in order to
        #            | produce a valid output, we will nevertheless dump, in
        #            | p_result, the replacements chars corresponding to the
        #            | missing ending delimiters. In that case, [toReopen] will
        #            | contain the list of these opening delimiters. This will
        #            | allow the caller be conscious of this fact, and, the next
        #            | time he will call us in order to dump a subsequent chunk
        #            | of p_text, he will possibly give us this list of tags
        #            | p_toReopen. Indeed, the real ending delimiters may be
        #            | present in the next p_text.
        # ----------------------------------------------------------------------
        # Ensure p_text is Unicode
        isStr = isinstance(text, str)
        if isStr: text = text.decode('utf-8')
        # Replace special markers when appropriate
        if fun: text = self.regex.sub(fun, text)
        # Build the result as a list
        r = []
        add = r.append
        repls = self.replacements
        delimiters = Replacements.delimiters
        endDelimiters = Replacements.endDelimiters
        previous = '' # The previous word or whitespace char
        # Start by dumping p_toReopen delimiters if any
        stack = toReopen # The stack of currently opened delimiters
        if toReopen:
            for delimiter in toReopen:
                add(self.getReplacementFor(delimiter, None))
        # Dump "normal" p_text
        for char in text:
            if char not in repls:
                # Leave the char untouched in the result
                add(char)
            else:
                # Apply a replacement
                add(self.getReplacementFor(char, previous))
                # Update the stack
                if char in delimiters:
                    if stack is None: stack = []
                    stack.append(char)
                elif (char in endDelimiters) and stack:
                    # We suppose p_text to be well structured and the last
                    # opened delimiter to correspond to this end delimiter. But
                    # we protect ourselves against possible mess and avoid
                    # popping the stack if empty.
                    stack.pop()
            # Update "previous"
            if (char in self.whitespace) or (previous in self.whitespace):
                previous = char
            else:
                previous += char
        # If the stack is not empty, dump corresponding ending delimiters
        if stack:
            i = len(stack) - 1
            while i >= 0:
                add(self.getReplacementFor(delimiters[stack[i]], None))
                i -= 1
        # Join the result to a string
        r = u''.join(r)
        # Encode the result when relevant
        if isStr: r = r.encode('utf-8')
        return r, stack

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Text2Html:
    '''Converts a chunk of structured text into XHTML. Some conventions apply,
       ie, rows of text starting with a dash are converted to bulleted lists;
       rows starting with char "|" are considered to be table rows, etc.'''
    defaultReplacements = Replacements()

    def __init__(self, p='p', prefix='', replacements='default',
                 replacementsFun=None, preListClass=None, lastLiClass=None):
        # The HTML tag used for representing a paragraph
        self.p = p
        # A chunk of HTML code that could be inserted just after the first
        # opening tag in the result.
        self.prefix = prefix
        # The result, will be a list that'll be joined at the end of the process
        self.res = None
        # If p_replacements is:
        # ----------------------------------------------------------------------
        #   "default"    | A default Replacements instance will be used, in
        #                | order to format text according to some conventions
        #                | (see class hereabove)
        # ----------------------------------------------------------------------
        #   None         | Text replacement is disabled
        # ----------------------------------------------------------------------
        #   an instance  | A Replacements instance containing a set of
        #                | formatting options that may de different from the
        #                | defaults.
        # ----------------------------------------------------------------------
        if replacements == 'default':
            self.replacements = Text2Html.defaultReplacements
        else:
            self.replacements = replacements
        # If a p_fun(ction) is defined, structured p_text can contain special
        # markers of the form
        #                            *<name>*
        # Evertytime such marker is encountered in the text, p_fun will be
        # called, with "name" as unique arg, and must return the replacement
        # text for this name.
        self.replacementsFun = replacementsFun
        # If "preListClass" is specified, it corresponds to a CSS class that
        # will be applied to the paragraph that is positioned just before a list
        # (ol or ul).
        self.preListClass = preListClass
        # If "lastLiClass" is specified, it corresponds to a CSS class that will
        # be applied to the last bullet in a list.
        self.lastLiClass = lastLiClass
        # The list of opened delimiters not closed after the last text converted
        # with replacements (see attribute Replacements.delimiters hereabove).
        self.opened = None

    def add(self, part, suffix='\n'):
        '''Adds some p_part into the result'''
        res = self.res
        wasEmpty = not res
        res.append(part + suffix)
        if wasEmpty and self.prefix:
            # It is time to dump the prefix if specified
            res.append(self.prefix)

    def addTag(self, tag, start=True):
        '''Adds a (starting or ending, depending on p_start) tag to the
           result.'''
        if tag in ('ul', 'ol'):
            # Surround the list by a paragraph. Else, self.prefix could not be
            # rendered.
            if start:
                self.add('<%s>' % self.p, '')
                self.add('<%s>' % tag)
            else:
                self.add('</%s>' % tag, '')
                self.add('</%s>' % self.p)
        else:
            t = not start and '/' or ''
            self.add('<%s%s>' % (t, tag))

    def convertText(self, text):
        '''Converts, within p_text, some chars according to self.replacements
           and return the result. Use p_self.replacementsFun if defined and
           p_useReplacementFun is True.'''
        if not self.replacements: return text
        r, opened = self.replacements.applyOn(text, fun=self.replacementsFun,
                                              toReopen=self.opened)
        # "opened" may contain the list of opened delimiters found in p_text
        # that were not closed within p_text. Remember them in p_self.opened: we
        # will re-open them in the next call to p_converText, because they were
        # automatically closed, in p_text, by m_applyOn hereabove.
        self.opened = opened
        return r

    def applyPreListClass(self):
        '''Apply the CSS class in self.preListClass to the last paragraph in
           self.res.'''
        r = self.res
        if not r: return
        i = len(r) - 1
        while i >= 0:
            if r[i].startswith('<p>'):
                r[i] = '<p class="%s"%s' % (self.preListClass, r[i][2:])
                break
            i -= 1

    def isLast(self, rows, i, elem):
        '''Is row #p_i within p_rows the last to contain some p_elem ?'''
        return (i+1 == len(rows)) or not rows[i+1].startswith(elem)

    def convert(self, s):
        '''Returns the converted chunk of HTML'''
        if not s: return self.prefix
        self.res = []
        inList = False
        inTable = False
        i = 0
        rows = s.split('\n')
        total = len(rows)
        for row in rows:
            if row.startswith('- '):
                # Convert it to a bullet. But are we already in a list ?
                if not inList:
                    # Apply the "prelist" CSS class to the last paragraph
                    if self.preListClass: self.applyPreListClass()
                    self.addTag('ul')
                    inList = True
                # Is this the last bullet ?
                isLast = self.isLast(rows, i, '- ')
                if isLast:
                    start= '<li class="%s">' % self.lastLiClass
                    end = '</li>'
                    inList = False
                else:
                    start = '<li>'
                    end = '</li>'
                self.add('%s%s%s' % (start, self.convertText(row[2:]), end))
                # Add the end list tag when relevant
                if isLast: self.addTag('ul', start=False)
            elif row.startswith('|'):
                # Convert it to a table row. But are we already in a table ?
                if not inTable:
                    self.addTag('table')
                    inTable = True
                # Split the row in cells
                tr = '<tr>'
                tdh = 'td'
                for cell in row.strip('|').split('|'):
                    if cell.startswith('_'):
                        tdh = 'th'
                        cell = cell[1:]
                    tr += '<%s>%s</%s>' % (tdh, self.convertText(cell), tdh)
                tr += '</tr>'
                self.add(tr)
                if self.isLast(rows, i, '|'):
                    self.add('</table>')
                    self.inTable = False
            else:
                self.addTag(self.p)
                # Allow the use of dashes that remain dashes (and are not
                # converted to bullets, as above).
                if row.startswith('-'): row = '- %s' % row[1:]
                self.add('%s</%s>' % (self.convertText(row), self.p))
            i += 1
        # Close the last opened list if any
        if inList: self.addTag('ul', start=False)
        return ''.join(self.res)

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Html2Text(Parser):
    '''Converts a chunk HTML into structured text'''

    # This parser uses conventions for default replacements only
    # ~~~
    # Conversion of HTML start tags
    startTags = {'b':'[', 'strong':'[', 'i':'<', 'em':'<', 'li':'- '}

    # Conversion of HTML end tags
    endTags = {'b':']', 'strong':']', 'i':'>', 'em':'>', 'p':'\n', 'li':'\n'}

    def startDocument(self):
        Parser.startDocument(self)
        self.r = []

    def endDocument(self):
        self.r = ''.join(self.r).strip()
        return Parser.endDocument(self)

    def characters(self, content):
        content = content.replace('\n', '').replace('\t', '')
        self.r.append(content)

    def startElement(self, elem, attrs):
        '''Convert HTML start tag p_elem to its text counterpart'''
        token = Html2Text.startTags.get(elem)
        if token: self.r.append(token) # else, ignore the tag

    def endElement(self, elem):
        '''Convert HTML end tag p_elem to its text counterpart'''
        token = Html2Text.endTags.get(elem)
        if token: self.r.append(token) # else, ignore the tag

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Icon:
    '''An icon from the toolbar when the Text field is used in "structured"
       mode.'''

    def __init__(self, name, type,
                 label=None, icon=None, data=None, shortcut=None):
        # A short, unique name for the icon
        self.name = name
        # The following type of icons exist. Depending on the type, p_data
        # carries a specific type of information.
        # ----------------------------------------------------------------------
        # p_type      | p_data
        # ----------------------------------------------------------------------
        # "wrapper"   | the icon corresponds to a portion of text that will be
        #             | wrapped around a start and end char. p_data contains 2
        #             | chars: the start and end wrapper chars.
        #             | 
        #             | For example, icon "bold" is of type "wrapper", with data
        #             | being "[]". When applied to selected text "hello", it
        #             | becomes "[hello]".
        # ----------------------------------------------------------------------
        # "char"      | the icon corresponds to a char to insert into the field.
        #             | p_data is the char to insert.
        # ----------------------------------------------------------------------
        # "action"    | the icon corresponds to some action that is not
        #             | necessarily related to the field content. In that case,
        #             | p_data may be None or its sematincs may be specific to
        #             | the action.
        # ----------------------------------------------------------------------
        # "sentences" | a clic on the icon will display a menu containing
        #             | predefined sentences. Selecting one of them will inject
        #             | it in the target field, where the cursor is currently
        #             | set. In that case, p_data must hold the name of a
        #             | method that must exist on the current object. This
        #             | method will be called without arg and must return a list
        #             | of sentences, each one being a string.
        # ----------------------------------------------------------------------
        self.type = type
        # The i18n label for the icon's tooltip. Should include the keyboard
        # shortcut when present. If None, defaults to "icon_<name>"
        self.label = label or ('icon_%s' % name)
        # The name of the icon image on disk. If None, will be computed as
        # "icon_<name>.png".
        self.icon = icon or ('icon_%s' % name)
        # The data related to this icon, as described hereabove
        self.data = data
        # If a keyboard shortcut is tied to the icon, its key code is defined
        # here, as an integer. See JavasScript keycodes, https://keycode.info.
        self.shortcut = shortcut

    def asSentences(self, r, o):
        '''For an icon of type "sentences", wraps the icon into a div allowing
           to hook the sub-div containing the sentences, and add this latter.'''
        # For an icon of type "sentences", add a div containing the sentences
        sentences = []
        for sentence in getattr(o, self.data)():
            if not isinstance(sentence, str):
                # We have an additional, custom info to add besides the sentence
                # itself.
                sentence, info = sentence
            else:
                info = ''
            div = '<div class="sentence"><a class="clickable" ' \
                  'onclick="injectSentence(this)" title="%s">%s</a>%s</div>' % \
                  (sentence, Px.truncateValue(sentence, width=65), info)
            sentences.append(div)
        # Add a warning message if no sentence has been found
        if not sentences:
            sentences.append('<div class="legend">%s</div>' % \
                             o.translate('no_sentence'))
        return '<div class="sentenceContainer" ' \
               'onmouseover="toggleDropdown(this) " ' \
               'onmouseout="toggleDropdown(this,\'none\')">%s' \
               '<div class="dropdown" style="display:none; width:350px">' \
               '%s</div></div>' % (r, '\n'.join(sentences))

    def get(self, o):
        '''Returns the HTML chunk representing this icon'''
        shortcut = str(self.shortcut) if self.shortcut else ''
        r = '<img class="icon" src="%s" title="%s" name="%s"' \
            ' onmouseover="switchIconBack(this, true)"' \
            ' onmouseout="switchIconBack(this, false)"' \
            ' data-type="%s" data-data="%s" data-shortcut="%d" ' \
            'onclick="useIcon(this)"/>' % \
             (o.buildUrl(self.icon), o.translate(self.label), self.name,
              self.type, self.data or '', shortcut)
        # Add specific stuff if icon type is "sentences"
        if self.type == 'sentences': r = self.asSentences(r, o)
        return r

# All available icons
Icon.all = [Icon('bold',      'wrapper', data='[]',       shortcut=66),
            Icon('italic',    'wrapper', data='&lt;&gt;', shortcut=73),
            Icon('highlight', 'wrapper', data='{}',       shortcut=72),
            # Non breaking space
            Icon('blank',     'char',    data=' ',        shortcut=32),
            # Non breaking dash
            Icon('dash',      'char',    data='‑',        shortcut=54),
            # Increment the field height by <data>%
            Icon('lengthen',  'action',  data='30',       shortcut=56)]

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Text(Multilingual, Field):
    '''Field allowing to encode a text made of several paragraphs, implemented
       by a HTML textarea tag.'''
    # Make some classes available here
    Icon = Icon
    Replacements = Replacements
    ToHtml = Text2Html
    FromHtml = Html2Text

    class Layouts(Layouts):
        '''Text-specific layouts'''
        b = Layouts(edit='lrv-d-f', view='l-f')
        g = Layouts(edit=Layout('d2-f;rv=', width='99%'),
                    view=Layout('fl', width='99%'))

        @classmethod
        def getDefault(class_, field):
            '''Default layouts for this Text p_field'''
            return class_.g if field.inGrid() else class_.b

    # Default ways to render the field if multilingual
    defaultLanguagesLayouts = {
      'edit': 'horizontal', 'view': 'vertical', 'cell': 'vertical'}

    # The name of the index class storing values of this field in the catalog
    indexType = 'TextIndex'

    # Unilingual view
    viewUni = cellUni = Px('''
     <x>::field.getInlineEditableValue(o, value or '-', layout, name=name,
                                       language=lg)</x>''')

    search = Px('''
     <input type="text" maxlength=":field.maxChars" size=":field.swidth"
            value=":field.sdefault" name=":widgetName"/>''')

    # The toolbar for structured content
    pxToolbar = Px('''
     <div class="toolbar" id=":tbid|field.name + '_tb'">
      <x for="icon in field.Icon.all">::icon.get(o)</x>
      <!-- Add inline-edition icons when relevant -->
      <x if="hostLayout">:field.pxInlineActions</x>
     </div>''',

     css = '''
      .toolbar { height: 24px; margin: 2px 0 }
      .sentenceContainer { position: relative; display: inline }
      .sentence { padding: 3px 0 }
      .icon { padding: 3px; border-width: 1px; border: 1px transparent solid }
      .iconSelected { background-color: #dbdbdb; border-color: #909090 }
     ''',

     js='''
      getIconsMapping = function(toolbar) {
        // Gets a mapping containing toolbar icons, keyed by their shortcut
        var r = {}, icons=toolbar.getElementsByClassName('icon');
        for (var i=0; i<icons.length; i++) {
          var icon=icons[i], key=icon.getAttribute('data-shortcut');
          if (key) r[parseInt(key)] = icon;
        }
        return r;
      }
      linkTextToolbar = function(toolbarId, target) {
        /* Link the toolbar with its target textarea. Get the target textarea if
           not given in p_target. */
        if (!target) {
          var targetId=_rsplit(toolbarId, '_', 2)[0];
          target = document.getElementById(targetId);
        }
        var toolbar=document.getElementById(toolbarId);
        toolbar['target'] = target;
        target['icons'] = getIconsMapping(toolbar);
      }
      switchIconBack = function(icon, selected) {
        icon.className = (selected)? 'icon iconSelected': 'icon';
      }
      lengthenArea = function(area, percentage) {
        // Lengthen some text p_area by some p_percentage
        var rate = 1 + (percentage / 100),
            styled = Boolean(area.style.height),
            height = (styled)? parseInt(area.style.height): area.rows;
        // Apply the rate
        height = Math.ceil(height * rate);
        // Reinject the new height to the correct area property
        if (styled) area.style.height = String(height) + 'px';
        else area.rows = height;
      }
      injectString = function(area, s) {
        // Inject some p_s(tring) into the text p_area, where the cursor is set
        var text = area.value,
                   start=area.selectionStart;
        area.value = text.substring(0, start) + s + \
                     text.substring(area.selectionEnd, area.value.length);
        area.selectionStart = area.selectionEnd = start +s.length;
        area.focus();
      }
      useIcon = function(icon) {
        // Get the linked textarea (if already linked)
        var area = icon.parentNode['target'];
        if (!area) return;
        var type=icon.getAttribute('data-type'),
            data=icon.getAttribute('data-data'),
            selectStart=area.selectionStart,
            selectEnd=area.selectionEnd,
            text=area.value;
        if (type == 'wrapper') {
          // Wrap the selected text within special chars
          area.value = text.substring(0, selectStart) + data[0] + \
                       text.substring(selectStart, selectEnd) + data[1] + \
                       text.substring(selectEnd, area.value.length);
          area.selectionStart = selectStart;
          area.selectionEnd = selectEnd + 2;
          area.focus();
        }
        else if (type == 'char') {
          // Insert a (sequence of) char(s) into the text
          injectString(area, data);
        }
        else if (type == 'action') {
          // Actions
          if (icon.name == 'lengthen') lengthenArea(area, parseInt(data));
        }
      }
      useShortcut = function(event, id) {
        if ((event.ctrlKey) && (event.keyCode in event.target['icons'])) {
          // Perform the icon's action
          useIcon(event.target['icons'][event.keyCode]);
          event.preventDefault();
        }
      }
      injectSentence = function(atag) {
        var area = atag.parentNode.parentNode.parentNode.parentNode['target'];
        if (!area) return;
        // Inject it
        injectString(area, atag.getAttribute('title'));
      }
     ''')

    # Buttons for saving or canceling while inline-editing the field. For a
    # structured text within a host layout, inline actions are rendered within
    # its toolbar.
    pxInlineActions = Px('''
      <div var="inToolbar=showToolbar and hostLayout;
                align='left' if inToolbar else 'right';
                fdir='row' if inToolbar else 'column'"
           style=":'float:%s;display:flex;flex-direction:%s' % (align, fdir)">
       <div><img id=":'%s_save' % tid" src=":url('save')"
                 class=":'clickable icon' if inToolbar else 'inlineIcon'"
                 title=":_('object_save')"/></div>
       <div><img id=":'%s_cancel' % tid" src=":url('cancel')"
                 class=":'clickable icon' if inToolbar else 'inlineIcon'"
                 title=":_('object_cancel')"/></div>
      </div>
      <script>:'prepareForAjaxSave(%s,%s,%s,%s)' % \
               (q(name), q(o.iid), q(o.url), q(hostLayout))</script>''')

    # Unilingual edit
    editUni = Px('''
     <x var="tid='%s_%s' % (name, lg) if lg else name;
             tbid='%s_tb' % tid;
             x=hostLayout and o.Lock.set(o, user, field=field);
             placeholder=field.getPlaceholder(o);
             showToolbar=field.showToolbar(ignoreInner=hostLayout)">

      <!-- Show the toolbar when relevant -->
      <x if="showToolbar">:field.pxToolbar</x>

      <!-- Add buttons for inline-edition when relevant -->
      <x if="not showToolbar and hostLayout">:field.pxInlineActions</x>

      <!-- The text zone in itself -->
      <textarea id=":tid" name=":tid" cols=":field.getTextareaCols()"
       style=":field.getTextareaStyle()" placeholder=":placeholder"
       readonly=":field.isReadonly(o)" onkeydown=":field.onKeyDown(tid)"
       onfocus=":field.onFocus(tid, lg, hostLayout)"
       rows=":field.height">:field.getInputValue(inRequest, requestValue, value)
      </textarea></x>''')

    pxSearch = Px('''
     <input type="text" maxlength=":field.maxChars" size=":field.swidth"
            value=":field.sdefault" name=":widgetName"/>''')

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, indexed=False, mustIndex=True, indexValue=None, searchable=False,
      filterField=None, readPermission='read', writePermission='write',
      width=60, height=5, maxChars=None, colspan=1, master=None,
      masterValue=None, focus=False, historized=False, mapping=None,
      generateLabel=None, label=None, sdefault='', scolspan=1, swidth=20,
      sheight=None, persist=True, inlineEdit=False, view=None, cell=None,
      edit=None, xml=None, translations=None,
      # Specific attributes
      placeholder=None, languages=('en',), languagesLayouts=None,
      structured=False, readonly=False):
        # You can define a placeholder in the following attribute. Please
        # consult the homonym attribute on class String from string.py for more
        # information.
        self.placeholder = placeholder
        # If attribute "structured" is True, the text encoded by your users in
        # the field will implicitly follow some conventions for applying text
        # formatting. A toolbar will be shown and will offer shortcuts allowing
        # to inject formatting marks (or also special chars) into the textarea.
        self.structured = structured
        # If attribute "readonly" is True (or stores a method returning True),
        # the rendered textarea tag, on edit layouts, will have attribute
        # "readonly" set.
        self.readonly = readonly
        # Call the base constructors
        Multilingual.__init__(self, languages, languagesLayouts)
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          None, searchable, filterField, readPermission, writePermission, width,
          height, maxChars, colspan, master, masterValue, focus, historized,
          mapping, generateLabel, label, sdefault, scolspan, swidth, sheight,
          persist, inlineEdit, view, cell, edit, xml, translations)
        # Specify a filter PX if the field content is indexed
        if self.indexed: self.filterPx = 'pxFilterText'
        self.checkParameters()

    def checkParameters(self):
        '''Prevent the use of some combinations of parameters'''
        # Currently, it is not possible to inline edit multilingual Texts
        langs = self.languages
        if (callable(langs) or (len(langs) > 1)) and self.inlineEdit:
            raise Exception(INLINE_EDIT_MULTILINGUAL)

    def getTextareaStyle(self):
        '''Get the content of textarea's "style" attribute'''
        # If the width is expressed as a string, the field width must be
        # expressed in attribute "style" (so returned by this method) and not
        # via attribute "cols" (returned by m_getTextareaCols below).
        return 'width:%s' % self.width if isinstance(self.width, str) else ''

    def getTextareaCols(self):
        '''Get the content of textarea's "cols" attribute'''
        # If width is expressed as an integer, it must be set as HTML attribute
        # "cols", and not as CSS attribute "width".
        return self.width if isinstance(self.width, int) else ''

    def showToolbar(self, ignoreInner=False):
        '''Show the toolbar if the field is structured and is not inner. In
           that latter case, the toolbar has already been rendered in the
           container field's headers.'''
        # Never show the tool bar if the field is not structured
        if not self.structured: return
        # Do not show the toolbar if the field is an inner field, provided this
        # check must be performed.
        return True if ignoreInner else not self.isInner()

    def onKeyDown(self, tid):
        '''When field is structured, this method returns the Javascript code to
           execute when a key is pressed, in order to implement keyboard
           shortcuts.'''
        return "useShortcut(event, '%s')" % tid if self.structured else ''

    def onFocus(self, tid, lg, hostLayout):
        '''When field is structured, this method returns the Javascript code to
           execute when the textarea gets focus, in order to link the field with
           the toolbar.'''
        if not self.structured: return ''
        if hostLayout:
            # We are inline-editing the (sub-)field: it has its own toolbar
            id = tid
        else:
            # For inner fields, there is a unique global toolbar
            id = '%s_%s' % (self.name, lg) if lg else self.name
        return "linkTextToolbar('%s_tb', this)" % id

    def getPlaceholder(self, o):
        '''Returns a placeholder for the field if defined'''
        r = self.getAttribute(o, 'placeholder') or ''
        if r == True:
            # A placeholder must be set, but we have no value. In this case, we
            # take the field label.
            r = o.translate(self.labelId)
        return r

    def getUniFormattedValue(self, o, value, layout='view', showChanges=False,
                             language=None, contentLanguage=None):
        '''Returns the formatted variant of p_value. If p_contentLanguage is
           specified, p_value is the p_contentLanguage part of a multilingual
           value.'''
        if Field.isEmptyValue(self, o, value) and not showChanges: return ''
        r = value
        if layout in ('view', 'cell'):
            r = Escape.xhtml(r, p=True)
        # If value starts with a carriage return, add a space; else, it will
        # be ignored.
        elif isinstance(r, str) and \
           (r.startswith('\n') or r.startswith('\r\n')): r = ' ' + r
        return r

    def getUniStorableValue(self, o, value):
        '''Gets the p_value as can be stored in the database within p_o'''
        if not value: return value
        # Clean the value
        value = value.replace('\r', '')
        # Manage maxChars
        max = self.maxChars
        if max and (len(value) > max): value = value[:max]
        return value

    def validateUniValue(self, o, value): return

    def getFilterValue(self, value):
        '''Ensure the value is suffixed with a star'''
        return value if value.endswith('*') else '%s*' % value

    @classmethod
    def computeSearchValue(class_, field, req, value=None):
        '''Converts text encoded in a search form into a range search or into
           individual words within an or-operator.'''
        r = Field.getSearchValue(field, req, value=value).strip()
        if not r: return r
        if r.endswith('*'):
            # Perform a "range" search
            r = r[:-1]
            if not r: return r
            r = r.split()[-1]
            r = in_(r, r+'z') # 'z' is higher than 'Z'
        else:
            # Perform a search based on keywords. No need to normalize it: it
            # will be done by the index afterwards.
            r = TextIndex.toIndexed(r, field)
            if not r:
                r = ''
            elif len(r) > 1:
                r = or_(*r)
            else:
                r = r[0]
        return r

    def getSearchValue(self, req, value=None):
        '''See called method's docstring'''
        return Text.computeSearchValue(self, req, value=value)

    def getListHeader(self, ctx):
        '''If this field is structured, when used as an inner field, the
           toolbar must be rendered only once, within the container field's
           header row corresponding to this field.'''
        # Inject the toolbar when appropriate
        if (ctx.layout == 'edit') and self.showToolbar(ignoreInner=True):
            bar = self.pxToolbar(ctx)
        else:
            bar = ''
        return '%s%s' % (Field.getListHeader(self, ctx), bar.encode('utf-8'))
#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
xhtmlInput = '''
<ol><li>
<p>Test du champ kupu<br />A la ligne
	1, j'y suis</p>
<ol><li>
<p>Ligne 1 de 1<br />A la ligne 1 de 1
		fdsfds fsd fsd fksdf sdfs dfd sfsd fsd fsd fdsf dsfds fsdfa azra
		zeeamr earkl kfl flks dlfksd lfklsd fklmsdfkl dskflmdsk flmsdf
		lmdsflm dflsdfs fksd fmlsd flmdsk flmdsf mlsfkmls<br />A la ligne 2
		de 1 cds fdsn sfd dsjfl dsjfhjds fhjdsf lqdhf klsfql kjfk jfkj
		qfklmqds fjdlksqfj kdlfj klqfk qfjk ljfqklq djfklqjf qfk jkfljqd
		sklfjqdklfjqdkl fjdqklf jdqlkfj qfq</p>
</li><li>
<p>Ligne 2 de 1<br />A la ligne 1 de 2
		fdsfds fsd fsd fksdf sdfs dfd sfsd fsd fsd fdsf dsfds fsdfa azra
		zeeamr earkl kfl flks dlfksd lfklsd fklmsdfkl dskflmdsk flmsdf
		lmdsflm dflsdfs fksd fmlsd flmdsk flmdsf mlsfkmls<br />A la ligne 2
		de 2 cds fdsn sfd dsjfl dsjfhjds fhjdsf lqdhf klsfql kjfk jfkj
		qfklmqds fjdlksqfj kdlfj klqfk qfjk ljfqklq djfklqjf qfk jkfljqd
		sklfjqdklfjqdkl fjdqklf jdqlkfj qf</p>
</li></ol>
</li><li>
<p>Ligne 2 tout court</p>
<ol><li>
<p>Ligne bullet dg fg dgd fgdf gdfg
		dgq fg fgfq gfqd gfqg qfg qgkqlglk lgkl fkgkfq lmgkl mfqfkglmfk
		gmlqf gmlqfgml kfmglk qmglk qmlgk qmlgkqmflgk qmlqg fmdlmslgk
		mlsgml fskfmglk gmlkflmg ksfmlgk mlsgk</p>
</li><li>
<p>dsfl kqfs dmflm dsfsdf lskfmls
		dkflmsdkf sdlmkf dslmfk sdmlfksd mlfksdmfl ksflmksdflmd slfmskd
		lsmlfk mlsdfkl mskfmlsfk lmskfsfs</p>
</li><li>
<p>fmlsdm ùfkùds fldsf ùsfsdmfù
		mdsfù msdùfms</p>
</li><li>
<p>fds fsd fdsf sdfds fsmd fmjdfklm
		sdflmkd lfqlmklmdsqkflmq dskflmkd slmgkqdfmglklfmgksmldgk
		dmlsgdkdlm fgkmdl fkgdmlfsgk mlfgksmdl fgkmldsf klmmdfkg mlsdfkgml
		skfdgml skgmlkfd smlgksd mlkgml kml</p>
</li><li>
<p>lgd ksmlgjk mlsdfgkml sdfkglm
		kdsfmlgk dlmskgsldmgk lms</p>
</li></ol>
</li><li>
<p>Ligne 3 tout court</p>
</li></ol>
<br />'''

xhtmlInput2 = '''
<ol start="1"><li>Le Gouvernement approuve la réaffectation de 20.056.989 € attribués dans le cadre du CRAC II et l’affectation du solde de 20.855.107 € du solde des CRAC Ibis et II au sein du secteur des hôpitaux de l’enveloppe de financement alternatif pour un montant total de 40.921.096 €, au bénéfice des établissements suivants :<br /><br /></li> 
<table align="center"> 
<tbody> 
<tr> 
<td> 
<b>Etablissement</b></td> 
<td> 
<p align="center" style="text-align: center;"><b>CRAC II Phase II</b></p> 
</td> 
</tr> 
<tr> 
<td> 
<p>C.H. Chrétien Liège</p> 
</td> 
<td nowrap="-1"> 
<p align="center" style="text-align: center;">11.097.377</p> 
</td> 
</tr> 
<tr> 
<td nowrap="-1"> 
<p>Hôp. St-Joseph, Ste-Thérèse et IMTR  Gilly</p> 
</td> 
<td nowrap="-1"> 
<p align="center" style="text-align: center;">8.297.880</p> 
</td> 
</tr>
</tbody> 
</table> 
<br /><li>Il prend acte des décisions du Ministre de la Santé relatives à l’octroi de la règle de 10/90 pour la subsidiation des infrastructures des établissements concernés.<br /></li>
<li>Le Gouvernement charge le Ministre de la Santé d’appliquer, le cas échéant, les nouveaux plafonds à la construction visés dans l’arrêté ministériel du 11 mai 2007 fixant le coût maximal pouvant être pris en considération pour l’octroi des subventions pour la construction de nouveaux bâtiments, les travaux d’extension et de reconditionnement d’un hôpital ou d’un service, aux demandes d’octroi de subventions antérieures au 1<sup>er</sup> janvier 2006, pour autant que ces demandes d’octroi de subventions n’aient pas encore donné lieu à l’exploitation à cette même date.<br /></li>
<li>Il charge le Ministre de la Santé de l’exécution de la présente décision</li></ol>
<p></p>
'''

xhtmlInput3 = '''
<ol><li>Le Gouvernement l'exercice 2008.</li><li>Il approuve 240.000€ de007-2008.</li><li>Le Gouvernement approuve: 
<ul><li>le projet d'arrêté ministériel 008;</li><li>le projet d'arrêté ministériel mique 2008-2009.</li></ul></li><li>Le Gouvernement charge le Ministre de l'Economie de l'exécution de la présente décision.</li></ol>
'''

xhtmlInput4 = '''
<div><strong>Programmes FSE Convergence et Compétitivité régionale et emploi.</strong></div>
<div><strong>Axe 1, mesure 1, et Axe 2, mesures 2, 3, 4 et 5 : formation.</strong></div>
<div><strong>Portefeuille de projets « Enseignement supérieur - Formation continue ».</strong></div>
'''



# -*- coding: utf-8 -*-
xhtmlInput1 = '''
<p><i><p>***<em>Vu les demandes</em></p>a</i></p>
'''

xhtmlInput2 = '''
<p><br/><b>Proposition(s) / <i>Voorstel(len)</i>:</b><p>Approuver</p>
<i><p>***<em>Vu les demandes</em></p>a</i></p>
'''

xhtmlInput3 = '''
<p><br/><b>Proposition(s) / <i>Voorstel(len)</i>:</b><p>Approuver l'octroi de concessions pour sépultures</p>
<p>Montant de la Recette : 5.340 € </p>
<p>Article budgétaire : 878/161-05/61</p>
<i><p>***<em>Vu les demandes tendant à obtenir des concessions de terrain pour sépulture au cimetière communal;</em></p>
<p><em>Vu l'article 123 de la loi communale;</em></p>
<p><em>Vu la loi du 20 juillet 1971 sur les funérailles et sépultures;</em></p>
<p><em>Vu le règlement communal sur les concessions de sépulture;</em></p>
<p><em>Vu la délibération du Conseil communal en date du 26 juin 1984 renouvelée le 16 octobre 2012 par laquelle celui-ci a fait usage de la faculté de délégation prévue par l'article 6 de la loi du 20 juillet 1971;</em></p>
<p><em>DECIDE :</em></p>
<p><em><u>Article 1er</u></em><em>: Les concessions reprises au tableau ci-annexé, numérotées du n° 83 au n° 90 inclus sont accordées aux conditions des règlements communaux sur la matière.</em></p>
<p><em><u>Article 2</u></em><em>: Un titre de concession sera remis au(x) requérant(s).</em></p>
<p>***<em>Gelet op de aanvragen tot het bekomen van grondvergunningen voor begravingen op de gemeentelijke begraafplaats;</em></p>
<p><em>Gelet op artikel 123 van de gemeentewet;</em></p>
<p><em>Gelet op de wet van 20 juli 1971 op de begraafplaatsen en de lijkbezorging;</em></p>
<p><em>Gelet op het gemeentelijk reglement betreffende de grafconcessies;</em></p>
<p><em>Gelet op het besluit van de Gemeenteraad van 26 juni 1984, vernieuwd op 16 oktober 2012 waarbij deze gebruik gemaakt heeft van het overdrachtsrecht, bepaald bij artikel 6 van de wet van 20 juli 1971;</em></p>
<p><em>BESLUIT :</em></p>
<p><em><u>Artikel 1</u></em><em>: De vergunningen hernomen op de hierbijgevoegde lijst, van nr 83 tot nr 90 inbegrepen verleend tegen voorwaarden bij de gemeentereglementen inzake bepaald.</em></p>
<p><em><u>Artikel 2</u></em><em>: Een vergunningstitel zal aan de aanvrager(s) overhandigd worden.</em></p>
</i><center style="font-weight: bold">Approuvé / Goedgekeurd</center><div class="itemRef">61/B/002</div></p>
'''

xhtmlInput4 = '''
<center style="font-weight: bold">Approuvé / Goedgekeurd</center>
'''



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.utils import string as sutils

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Batch:
    '''Represents a list of objects being part of a wider list, ie:
        * page 2 displaying objects 30 to 60 from a search matching 1110
          results;
        * page 1 displaying objects 0 to 20 from a list of 43 referred objects.
    '''
    def __init__(self, objects=None, total=0, size=30, start=0, hook=None):
        # The objects being part of this batch
        self.objects = objects
        # The effective number of objects in this batch
        self.length = len(objects) if objects is not None else 0
        # The total number of objects from which this batch is only a part of
        self.total = total
        # The normal size of a batch. If p_size is None, all objects are shown.
        # "length" can be less than "size". Example: if total is 12 and batch
        # size is 10, the batch representing objects 0 to 9 will have
        # length=size=10, but the second batch representing objects 10 to 12
        # will have size (3) < length (10).
        self.size = size
        # The index of the first object in the current batch
        self.start = start
        # When this batch is shown in the ui, the ID of the DOM node containing
        # the list of objects.
        self.hook = hook

    def setObjects(self, objects):
        '''p_objects may no have been added at construction time'''
        self.objects = objects
        self.length = len(objects)

    # Input field for going to element number x
    pxGotoNumber = Px('''
     <x var2="label=_('goto_number');
              gotoName='%d_%s_goto' % (o.iid, field.name)">
      <span style="padding-left: 5px">:label</span>
      <input type="text" size=":len(str(total)) or 1" onclick="this.select()"
             onkeydown=":'if (event.keyCode==13) document.getElementById' \
                         '(%s).click()' % q(gotoName)"/><img
             id=":gotoName" name=":gotoName"
             class="clickable" src=":url('gotoNumber')" title=":label"
             onclick=":'gotoTied(%s,%s,this.previousSibling,%s,%s)' % \
                 (q(sourceUrl), q(field.name), total, q(popup))"/></x>''')

    pxNavigate = Px('''
     <x if="batch.total &gt; batch.size"
        var2="hook=q(batch.hook); size=q(batch.size)">

      <!-- Go to the first page -->
      <img if="(batch.start != 0) and (batch.start != batch.size)"
           class="clickable iconS" src=":url('arrows.svg')"
           title=":_('goto_first')" style="transform: rotate(90deg)"
           onclick=":'askBunch(%s,%s,%s)'% (hook, q(0), size)"/>

      <!-- Go to the previous page -->
      <img var="sNumber=batch.start - batch.size" if="batch.start != 0"
           class="clickable iconS" src=":url('arrow.svg')"
           title=":_('goto_previous')" style="transform: rotate(90deg)"
           onclick=":'askBunch(%s,%s,%s)' % (hook, q(sNumber), size)"/>

      <!-- Explain which elements are currently shown -->
      <span class="navText"> 
       <x>:batch.start + 1</x> <img src=":url('to')"/> 
       <x>:batch.start + batch.length</x> <span class="navSep">//</span> 
       <x>:batch.total</x>
      </span>

      <!-- Go to the next page -->
      <img var="sNumber=batch.start + batch.size" if="sNumber &lt; batch.total"
           class="clickable iconS" src=":url('arrow.svg')"
           title=":_('goto_next')" style="transform: rotate(270deg)"
           onclick=":'askBunch(%s,%s,%s)' % (hook, q(sNumber), size)"/>

      <!-- Go to the last page -->
      <img var="lastPageIsIncomplete=batch.total % batch.size;
                nbOfCompletePages=int(batch.total / batch.size);
                nbOfCountedPages=lastPageIsIncomplete and \
                                 nbOfCompletePages or nbOfCompletePages-1;
                sNumber= nbOfCountedPages * batch.size"
           if="(batch.start != sNumber) and \
               (batch.start != (sNumber-batch.size))"
           class="clickable iconS" src=":url('arrows.svg')"
           title=":_('goto_last')" style="transform: rotate(270deg)"
           onclick=":'askBunch(%s,%s,%s)' % (hook, q(sNumber), size)"/>

      <!-- Go to the element number... -->
      <x var="gotoNumber=gotoNumber|False" if="gotoNumber"
         var2="sourceUrl=o.url; total=batch.total">:batch.pxGotoNumber</x>
     </x>''')

    def store(self, search):
        '''Returns the Javascript code allowing to store objects from this batch
           in the browser's session storage.'''
        # Get the key at which we will store object IDs in the local storage
        key = search.getSessionKey()
        # Get the dict of IDs to store ~{i_index:i_id}~
        i = self.start
        value = {}
        for o in self.objects:
            value[i] = o.iid
            i += 1
        value = sutils.getStringFrom(value)
        return "sessionStorage.setItem('%s',JSON.stringify(%s))" % (key, value)

    def __repr__(self):
        '''String representation'''
        data = 'start=%d,length=%d,size=%s,total=%d' % \
               (self.start, self.length, self.size, self.total)
        if self.hook: data = 'hook=%s,%s' % (self.hook, data)
        return '<Batch %s>' % data
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import time

from appy.px import Px
from appy.fields import Field
from appy.shared import dates
from appy.fields.hour import Hour

# ------------------------------------------------------------------------------
def getDateFromIndexValue(indexValue):
    '''p_indexValue is the internal representation of a date as stored in the
       zope Date index (see "_convert" method in DateIndex.py in
       Products.pluginIndexes/DateIndex). This function produces a DateTime
       based on it.'''
    # p_indexValue represents a number of minutes
    minutes = indexValue % 60
    indexValue = (indexValue-minutes) / 60 # The remaining part, in hours
    # Get hours
    hours = indexValue % 24
    indexValue = (indexValue-hours) / 24 # The remaining part, in days
    # Get days
    day = indexValue % 31
    if day == 0: day = 31
    indexValue = (indexValue-day) / 31 # The remaining part, in months
    # Get months
    month = indexValue % 12
    if month == 0: month = 12
    year = (indexValue - month) / 12
    from DateTime import DateTime
    utcDate = DateTime('%d/%d/%d %d:%d UTC' % (year,month,day,hours,minutes))
    return utcDate.toZone(utcDate.localZone())

# ------------------------------------------------------------------------------
class Date(Field):
    pxView = pxCell = Px('''<x>:value</x>''')

    # PX for selecting hour and minutes, kindly provided by field Hour
    pxHour = Hour.pxEdit

    pxEdit = Px('''
     <x var="years=field.getSelectableYears(zobj)">
      <!-- Day -->
      <select if="field.showDay" var2="days=range(1,32); part='%s_day' % name"
              name=":part" id=":part">
       <option value="">-</option>
       <option for="day in days" var2="zDay=str(day).zfill(2)" value=":zDay"
         selected=":field.isSelected(zobj, part, 'day', \
                                     day, rawValue)">:zDay</option>
      </select> 

      <!-- Month -->
      <select var="months=range(1,13); part='%s_month' % name"
              name=":part" id=":part">
       <option value="">-</option>
       <option for="month in months"
         var2="zMonth=str(month).zfill(2)" value=":zMonth"
         selected=":field.isSelected(zobj, part, 'month', \
                                     month, rawValue)">:zMonth</option>
      </select> 

      <!-- Year -->
      <select var="part='%s_year' % name" name=":part" id=":part">
       <option value="">-</option>
       <option for="year in years" value=":year"
         selected=":field.isSelected(zobj, part, 'year', \
                                     year, rawValue)">:year</option>
      </select>

      <!-- The icon for displaying the calendar popup -->
      <x if="field.calendar">
       <input type="hidden" id=":name" name=":name"/>
       <img id=":'%s_img' % name" src=":url('calendar.gif')"/>
       <script type="text/javascript">::field.getJsInit(name, years)</script>
      </x>

      <!-- Hour and minutes -->
      <x if="field.format == 0">:field.pxHour</x>
     </x>''')

    pxSearch = Px('''
     <table var="years=field.getSelectableYears(zobj);
                 dstart,dend=field.getDefaultSearchValues(zobj)">
       <!-- From -->
       <tr var="fromName='%s_from' % name;
                dayFromName='%s_from_day' % name;
                monthFromName='%s_from_month' % name">
        <td width="10px">&nbsp;</td>
        <td><label>:_('search_from')</label></td>
        <td>
         <select id=":dayFromName" name=":dayFromName">
          <option value="">--</option>
          <option for="value in [str(v).zfill(2) for v in range(1, 32)]"
                  value=":value" selected=":value == dstart[2]">:value</option>
         </select> / 
         <select id=":monthFromName" name=":monthFromName">
          <option value="">--</option>
          <option for="value in [str(v).zfill(2) for v in range(1, 13)]"
                  value=":value" selected=":value == dstart[1]">:value</option>
         </select> / 
         <select id=":widgetName" name=":widgetName">
          <option value="">--</option>
          <option for="value in years"
                  value=":value" selected=":value == dstart[0]">:value</option>
         </select>
         <!-- The icon for displaying the calendar popup -->
         <x if="field.calendar">
          <input type="hidden" id=":fromName" name=":fromName"/>
          <img id=":'%s_img' % fromName" src=":url('calendar.gif')"/>
          <script type="text/javascript">::field.getJsInit(fromName, years)
          </script>
         </x>
         <!-- Hour and minutes when relevant -->
         <x if="(field.format == 0) and field.searchHour"
            var2="hPart='%s_from_hour' % name;
                  mPart='%s_from_minute' % name">:field.pxHour</x>
        </td>
       </tr>

       <!-- To -->
       <tr var="toName='%s_to' % name;
                dayToName='%s_to_day' % name;
                monthToName='%s_to_month' % name;
                yearToName='%s_to_year' % name">
        <td></td>
        <td><label>:_('search_to')</label>&nbsp;&nbsp;&nbsp;&nbsp;</td>
        <td height="20px">
         <select id=":dayToName" name=":dayToName">
          <option value="">--</option>
          <option for="value in [str(v).zfill(2) for v in range(1, 32)]"
                  value=":value" selected=":value == dend[2]">:value</option>
         </select> / 
         <select id=":monthToName" name=":monthToName">
          <option value="">--</option>
          <option for="value in [str(v).zfill(2) for v in range(1, 13)]"
                  value=":value" selected=":value == dend[1]">:value</option>
         </select> / 
         <select id=":yearToName" name=":yearToName">
          <option value="">--</option>
          <option for="value in years"
                  value=":value" selected=":value == dend[0]">:value</option>
         </select>
         <!-- The icon for displaying the calendar popup -->
         <x if="field.calendar">
          <input type="hidden" id=":toName" name=":toName"/>
          <img id=":'%s_img' % toName" src=":url('calendar.gif')"/>
          <script type="text/javascript">::field.getJsInit(toName, years)
          </script>
         </x>
         <!-- Hour and minutes when relevant -->
         <x if="(field.format == 0) and field.searchHour"
            var2="hPart='%s_to_hour' % name;
                  mPart='%s_to_minute' % name">:field.pxHour</x>
        </td>
       </tr>
      </table>''')

    # Required CSS and Javascript files for this type
    cssFiles = {'edit': ('jscalendar/calendar-blue.css',)}
    jsFiles = {'edit': ('jscalendar/calendar.js',
                        'jscalendar/lang/calendar-en.js',
                        'jscalendar/calendar-setup.js')}
    # Possible values for "format"
    WITH_HOUR = 0
    WITHOUT_HOUR = 1
    dateParts = ('year', 'month', 'day')
    hourParts = ('hour', 'minute')
    # Default value on the search screen
    searchDefault = (None, None, None)

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, format=WITH_HOUR, dateFormat=None, hourFormat=None,
      calendar=True, startYear=time.localtime()[0]-10,
      endYear=time.localtime()[0]+10, reverseYears=False, minutesPrecision=5,
      show=True, page='main', group=None, layouts=None, move=0, indexed=False,
      mustIndex=True, indexValue=None, searchable=False,
      specificReadPermission=False, specificWritePermission=False, width=None,
      height=None, maxChars=None, colspan=1, master=None, masterValue=None,
      focus=False, historized=False, mapping=None, generateLabel=None,
      label=None, sdefault=None, scolspan=1, swidth=None, sheight=None,
      persist=True, view=None, cell=None, edit=None, xml=None,
      translations=None, showDay=True, searchHour=False):
        self.format = format
        self.calendar = calendar
        self.startYear = startYear
        self.endYear = endYear
        # If reverseYears is True, in the selection box, available years, from
        # self.startYear to self.endYear will be listed in reverse order.
        self.reverseYears = reverseYears
        # If p_showDay is False, the list for choosing a day will be hidden
        self.showDay = showDay
        # If no p_dateFormat/p_hourFormat is specified, the application-wide
        # tool.dateFormat/tool.hourFormat is used instead.
        self.dateFormat = dateFormat
        self.hourFormat = hourFormat
        # In the context of a Date, the max hour is always 23. But it can be
        # more in the context of an Hour field.
        self.maxHour = 23
        # If "minutesPrecision" is 5, only a multiple of 5 can be encoded. If
        # you want to let users choose any number from 0 to 59, set it to 1.
        self.minutesPrecision = minutesPrecision
        # The search widget will only allow to specify start and end dates
        # without hour, event if format is WITH_HOUR, excepted if searchHour is
        # True.
        self.searchHour = searchHour
        # Value for p_sdefault must be a tuple (start, end), where each value
        # ("start" or "end"), if not None, must be a tuple (year, month, day).
        # Each of these sub-values can be None or an integer value. p_sdefault
        # can also be a method that produces the values in the specified format.
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          searchable, specificReadPermission, specificWritePermission, width,
          height, None, colspan, master, masterValue, focus, historized,
          mapping, generateLabel, label, sdefault, scolspan, swidth, sheight,
          persist, False, view, cell, edit, xml, translations)

    def getCss(self, layoutType, res, config):
        # CSS files are only required if the calendar must be shown
        if self.calendar: Field.getCss(self, layoutType, res, config)

    def getJs(self, layoutType, res, config):
        # Javascript files are only required if the calendar must be shown
        if self.calendar: Field.getJs(self, layoutType, res, config)

    def getSelectableYears(self, obj):
        '''Gets the list of years one may select for this field'''
        startYear = self.getAttribute(obj, 'startYear')
        r = range(startYear, self.endYear + 1)
        if self.reverseYears: r.reverse()
        return r

    def validateValue(self, obj, value):
        DateTime = obj.getProductConfig().DateTime
        try:
            value = DateTime(value)
        except DateTime.DateError, ValueError:
            return obj.translate('bad_date')

    def getFormattedValue(self, obj, value, layoutType='view',
                          showChanges=False, language=None):
        if self.isEmptyValue(obj, value): return ''
        # Get the applicable date format
        tool = obj.getTool().appy()
        dateFormat = self.dateFormat or tool.dateFormat
        # A problem may occur with some extreme year values. Replace the "year"
        # part "by hand".
        if '%Y' in dateFormat:
            dateFormat = dateFormat.replace('%Y', str(value.year()))
        res = tool.o.formatDate(value, dateFormat, withHour=False)
        if self.format == Date.WITH_HOUR:
            res += ' %s' % value.strftime(self.hourFormat or tool.hourFormat)
        return res

    def getRequestValue(self, obj, requestName=None):
        request = obj.REQUEST
        name = requestName or self.name
        # Manage the "date" part
        value = ''
        for part in self.dateParts:
            # The "day" part may be hidden. Use "1" by default.
            if (part == 'day') and not self.showDay:
                valuePart = '01'
            else:
                valuePart = request.get('%s_%s' % (name, part), None)
            if not valuePart: return
            value += valuePart + '/'
        value = value[:-1]
        # Manage the "hour" part
        if self.format == self.WITH_HOUR:
            value += ' '
            for part in self.hourParts:
                valuePart = request.get('%s_%s' % (name, part), None)
                if not valuePart: return
                value += valuePart + ':'
            value = value[:-1]
        return value

    def searchValueIsEmpty(self, form):
        '''We consider a search value being empty if both "from" and "to" values
           are empty. At an individual level, a "from" or "to" value is
           considered not empty if at least the year is specified.'''
        # The base method determines if the "from" year is empty
        isEmpty = Field.searchValueIsEmpty
        return isEmpty(self, form) and \
               isEmpty(self, form, widgetName='%s_to_year' % self.name)

    def getRequestSuffix(self): return '_year'

    def getStorableValue(self, obj, value, complete=False):
        if not self.isEmptyValue(obj, value):
            import DateTime
            return DateTime.DateTime(value)

    def getDateFromSearchValue(self, year, month, day, hour, setMin):
        '''Gets a valid DateTime instance from date information coming from the
           request as strings in p_year, p_month and p_day. Returns None if
           p_year is empty. If p_setMin is True, when some information is
           missing (month or day), we will replace it with the minimum value
           (=1). Else, we will replace it with the maximum value (=12, =31).'''
        if not year: return
        # Set month and day
        if not month:
            month = setMin and 1 or 12
        if not day:
            day = setMin and 1 or 31
        # Set the hour
        if hour is None:
            hour = setMin and '00:00' or '23:59'
        # The specified date may be invalid (ie, 2018/02/31): ensure to produce
        # a valid date in all cases.
        from DateTime import DateTime
        try:
            r = DateTime('%s/%s/%s %s' % (year, month, day, hour))
        except:
            base = DateTime('%s/%s/01' % (year, month))
            r = dates.getLastDayOfMonth(base, hour=hour)
        return r

    def getSearchValue(self, form):
        '''Converts the raw search values from p_form into an interval of
           dates.'''
        # Get the "from" value
        name = self.name
        year = Field.getSearchValue(self, form)
        month = form['%s_from_month' % name]
        day   = form['%s_from_day' % name]
        hour = None
        if self.searchHour:
            hour = '%s:%s' % (form['%s_from_hour' % name] or '00',
                              form['%s_from_minute' % name] or '00')
        fromDate = self.getDateFromSearchValue(year, month, day, hour, True)
        # Get the "to" value"
        year  = form['%s_to_year' % name]
        month = form['%s_to_month' % name]
        day   = form['%s_to_day' % name]
        hour = None
        if self.searchHour:
            hour = '%s:%s' % (form['%s_to_hour' % name] or '23',
                              form['%s_to_minute' % name] or '59')
        toDate = self.getDateFromSearchValue(year, month, day, hour, False)
        return fromDate, toDate

    def getDefaultSearchValues(self, obj):
        '''Gets the default value for this field when shown on a search
           layout.'''
        default = self.getAttribute(obj, 'sdefault')
        if not default:
            r = self.searchDefault, self.searchDefault
        else:
            # Convert months and days to zfilled strings
            r = []
            for i in (0, 1):
                value = default[i]
                if value:
                    year, month, day = value
                    if month is not None: month = str(month).zfill(2)
                    if day is not None: day = str(day).zfill(2)
                    value = (year, month, day)
                else:
                    value = self.searchDefault
                r.append(value)
        return r

    def getIndexType(self): return 'DateIndex'

    def isSelected(self, obj, part, fieldPart, dateValue, dbValue):
        '''When displaying this field, must the particular p_dateValue be
           selected in the sub-field p_fieldPart corresponding to the date
           part?'''
        # Get the value we must compare (from request or from database)
        rq = obj.REQUEST
        if rq.has_key(part):
            compValue = rq.get(part)
            if compValue.isdigit():
                compValue = int(compValue)
        else:
            compValue = dbValue
            if compValue:
                compValue = getattr(compValue, fieldPart)()
        # Compare the value
        return compValue == dateValue

    def isSortable(self, usage):
        '''Can this field be sortable ?'''
        if usage == 'search': return Field.isSortable(self, usage)
        return True # Sortable in Ref fields

    def getJsInit(self, name, years):
        '''Gets the Javascript init code for displaying a calendar popup for
           this field, for an input named p_name (which can be different from
           self.name if, ie, it is a search field).'''
        # Always express the range of years in chronological order.
        years = [years[0], years[-1]]
        years.sort()
        return 'Calendar.setup({inputField: "%s", button: "%s_img", ' \
               'onSelect: onSelectDate, range:%s, firstDay: 1})' % \
               (name, name, str(years))
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
import os, os.path, sys
try:
    from AccessControl.SecurityManagement import \
         newSecurityManager, noSecurityManager
except ImportError:
    pass

# ------------------------------------------------------------------------------
class TestMixin:
    '''This class is mixed in with any ZopeTestCase.'''
    def getNonEmptySubModules(self, moduleName):
        '''Returns the list of sub-modules of p_app that are non-empty.'''
        res = []
        try:
            exec 'import %s' % moduleName
            exec 'moduleObj = %s' % moduleName
            moduleFile = moduleObj.__file__
            if moduleFile.endswith('.pyc'):
                moduleFile = moduleFile[:-1]
        except ImportError, ie:
            return res
        except SyntaxError, se:
            return res
        # Include the module if not empty. "Emptyness" is determined by the
        # absence of names beginning with other chars than "__".
        for elem in moduleObj.__dict__.iterkeys():
            if not elem.startswith('__'):
                res.append(moduleObj)
                break
        # Include sub-modules if any
        if moduleFile.find("__init__.py") != -1:
            # Potentially, sub-modules exist.
            moduleFolder = os.path.dirname(moduleFile)
            for elem in os.listdir(moduleFolder):
                if elem.startswith('.'): continue
                subModuleName, ext = os.path.splitext(elem)
                if ((ext == '.py') and (subModuleName != '__init__')) or \
                   os.path.isdir(os.path.join(moduleFolder, subModuleName)):
                    # Submodules may be sub-folders or Python files
                    subModuleName = '%s.%s' % (moduleName, subModuleName)
                    res += self.getNonEmptySubModules(subModuleName)
        return res

    @staticmethod
    def getCovFolder():
        '''Returns the folder where to put the coverage folder if needed.'''
        for arg in sys.argv:
            if arg.startswith('[coverage'):
                return arg[10:].strip(']')
        return
    def _setup(self): pass

# Functions executed before and after every test -------------------------------
def beforeTest(test):
    '''Is executed before every test.'''
    g = test.globs
    g['tool'] = test.app.config.appy()
    cfg = g['tool'].o.getProductConfig()
    g['appFolder'] = cfg.diskFolder
    moduleOrClassName = g['test'].name # Not used yet.
    # Initialize the test
    g['t'] = g['test']

def afterTest(test):
    '''Is executed after every test.'''
    g = test.globs
    appName = g['tool'].o.getAppName()
    exec 'from Products.%s import cov, covFolder, totalNumberOfTests, ' \
         'countTest' % appName
    countTest()
    exec 'from Products.%s import numberOfExecutedTests' % appName
    if cov and (numberOfExecutedTests == totalNumberOfTests):
        cov.stop()
        appModules = test.getNonEmptySubModules(appName)
        # Dumps the coverage report
        # HTML version
        cov.html_report(directory=covFolder, morfs=appModules)
        # Summary in a text file
        f = file('%s/summary.txt' % covFolder, 'w')
        cov.report(file=f, morfs=appModules)
        f.close()
        # Annotated modules
        cov.annotate(directory=covFolder, morfs=appModules)
# ------------------------------------------------------------------------------



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import os, os.path, sys, zipfile, re, shutil
from appy.shared.test import Tester, TestFactory, TesterError
from appy.shared.test import Test as BaseTest
from appy.shared.utils import FolderDeleter
from appy.shared.xml_parser import Escape
from appy.pod.odf_parser import OdfEnvironment, OdfParser
from appy.pod.renderer import Renderer
from appy.pod.styles_manager import \
     TableProperties, BulletedProperties, NumberedProperties

# TesterError-related constants ------------------------------------------------
TEMPLATE_NOT_FOUND = 'Template file "%s" was not found.'
CONTEXT_NOT_FOUND = 'Context file "%s" was not found.'
EXPECTED_RESULT_NOT_FOUND = 'Expected result "%s" was not found.'

# ------------------------------------------------------------------------------
def sjoin(folder, name):
    '''Shorthand for os.path.joining unicode strings.'''
    return str(os.path.join(folder, name))

# ------------------------------------------------------------------------------
class AnnotationsRemover(OdfParser):
    '''This parser is used to remove from content.xml and styles.xml the
       Python tracebacks that may be dumped into OpenDocument annotations by
       pod when generating errors. Indeed, those tracebacks contain lot of
       machine-specific info, like absolute paths to the python files, etc.'''
    def __init__(self, env, caller):
        OdfParser.__init__(self, env, caller)
        self.res = u''
        self.inAnnotation = False # Are we parsing an annotation ?
        self.textEncountered = False # Within an annotation, have we already
        # met a text ?
        self.ignore = False # Must we avoid dumping the current tag/content
        # into the result ?

    def startElement(self, elem, attrs):
        e = OdfParser.startElement(self, elem, attrs)
        # Do we enter into an annotation ?
        if elem == '%s:annotation' % e.ns(e.NS_OFFICE):
            self.inAnnotation = True
            self.textEncountered = False
        elif elem == '%s:p' % e.ns(e.NS_TEXT):
            if self.inAnnotation:
                if not self.textEncountered:
                    self.textEncountered = True
                else:
                    self.ignore = True
        if not self.ignore:
            self.res += '<%s' % elem
            for attrName, attrValue in attrs.items():
                self.res += ' %s="%s"' % \
                            (attrName, Escape.xml(attrValue, escapeApos=True))
            self.res += '>'

    def endElement(self, elem):
        e = self.env
        if elem == '%s:annotation' % e.ns(e.NS_OFFICE):
            self.inAnnotation = False
            self.ignore = False
        if not self.ignore:
            self.res += '</%s>' % elem
        OdfParser.endElement(self, elem)

    def characters(self, content):
        e = OdfParser.characters(self, content)
        if not self.ignore: self.res += Escape.xml(content)

    def getResult(self):
        return self.res

# ------------------------------------------------------------------------------
class Test(BaseTest):
    '''Abstract test class'''
    interestingOdtContent = ('content.xml', 'styles.xml')
    # "list-style"s can be generated in random order
    ignoreTags = ((OdfEnvironment.NS_DC, 'date'),
                  (OdfEnvironment.NS_STYLE, 'style'),
                  (OdfEnvironment.NS_TEXT, 'list-style'),
                  (OdfEnvironment.NS_STYLE, 'font-face'))
    # 'text:style-name's can contained generated names base on time.time
    ignoreAttrs = ('draw:name', 'text:name', 'text:bullet-char', 'table:name',
                   'table:style-name', 'text:style-name', 'xlink:href',
                   'xml:id')

    def __init__(self, testData, testDescription, testFolder,
                 config, flavour, rendererParams):
        BaseTest.__init__(self, testData, testDescription, testFolder, config,
                          flavour, rendererParams)
        self.templatesFolder = os.path.join(self.testFolder, 'templates')
        self.contextsFolder = os.path.join(self.testFolder, 'contexts')
        self.resultsFolder = os.path.join(self.testFolder, 'results')
        self.result = None
        self.rendererParams = rendererParams

    def getContext(self, contextName):
        '''Gets the objects that are in the context'''
        contextPy = os.path.join(self.contextsFolder, contextName + '.py')
        if not os.path.exists(contextPy):
            raise TesterError(CONTEXT_NOT_FOUND % contextPy)
        contextPkg = 'appy.pod.test.contexts.%s' % contextName
        exec 'import %s' % contextPkg
        exec 'context = dir(%s)' % contextPkg
        res = {}
        for elem in context:
            if not elem.startswith('__'):
                exec 'res[elem] = %s.%s' % (contextPkg, elem)
        return res

    def do(self):
        tempFileName = '%s.%s' % (self.data['Name'], self.data['Result'])
        self.result = sjoin(self.tempFolder, tempFileName)
        # Get the path to the template to use for this test
        if self.data['Template'].endswith('.ods'):
            suffix = ''
        else:
            # For ODT, which is the most frequent case, no need to specify the
            # file extension.
            suffix = '.odt'
        template = sjoin(self.templatesFolder, self.data['Template'] + suffix)
        if not os.path.exists(template):
            raise TesterError(TEMPLATE_NOT_FOUND % template)
        # Get the context
        context = self.getContext(self.data['Context'])
        # Get the LibreOffice port
        ooPort = self.data['LibreOfficePort']
        pythonWithUno = self.config['pythonWithUnoPath']
        # Get the styles mapping. Dicts are not yet managed by the TablesParser
        stylesMapping = eval('{' + self.data['StylesMapping'] + '}')
        # Call the renderer
        params = {'ooPort': ooPort, 'pythonWithUnoPath': pythonWithUno,
                  'stylesMapping': stylesMapping, 'protection': True}
        params.update(self.rendererParams)
        Renderer(template, context, self.result, **params).run()

    def getOdtContent(self, odtFile):
        '''Creates in the temp folder content.xml and styles.xml extracted
           from p_odtFile.'''
        contentXml = None
        stylesXml = None
        if odtFile == self.result:
            filePrefix = 'actual'
        else:
            filePrefix = 'expected'
        zipFile = zipfile.ZipFile(odtFile)
        for zippedFile in zipFile.namelist():
            if zippedFile in self.interestingOdtContent:
                f = file(os.path.join(self.tempFolder,
                                      '%s.%s' % (filePrefix, zippedFile)), 'wb')
                fileContent = zipFile.read(zippedFile)
                # Python tracebacks that are in annotations include the full
                # path to the Python files, which of course may be different
                # from one machine to the other. So we remove those paths.
                remover = AnnotationsRemover(OdfEnvironment(), self)
                remover.parse(fileContent)
                fileContent = remover.getResult()
                try:
                    f.write(fileContent.encode('utf-8'))
                except UnicodeDecodeError:
                    f.write(fileContent)
                f.close()
        zipFile.close()

    def checkResult(self):
        '''r_ is False if the test succeeded'''
        # Get styles.xml and content.xml from the actual result
        res = False
        self.getOdtContent(self.result)
        # Get styles.xml and content.xml from the expected result
        expectedResult = os.path.join(self.resultsFolder,
                                  self.data['Name'] + '.' + self.data['Result'])
        if not os.path.exists(expectedResult):
            raise TesterError(EXPECTED_RESULT_NOT_FOUND % expectedResult)
        self.getOdtContent(expectedResult)
        for fileName in self.interestingOdtContent:
            diffOccurred = self.compareFiles(
                os.path.join(self.tempFolder, 'actual.%s' % fileName),
                os.path.join(self.tempFolder, 'expected.%s' % fileName),
                areXml=True, xmlTagsToIgnore=Test.ignoreTags,
                xmlAttrsToIgnore=Test.ignoreAttrs, encoding='utf-8')
            if diffOccurred:
                res = True
                break
        return res

# Concrete test classes --------------------------------------------------------
class NominalTest(Test): pass
class ErrorTest(Test):
    def onError(self):
        '''Compares the error that occurred with the expected error'''
        Test.onError(self)
        return not self.isExpectedError(self.data['Message'])

# ------------------------------------------------------------------------------
class PodTestFactory(TestFactory):
    @staticmethod
    def createTest(testData, testDescription, testFolder, config, flavour,
                   rendererParams):
        isErrorTest = testData.table.instanceOf('ErrorTest')
        testClass = isErrorTest and ErrorTest or NominalTest
        return testClass(testData, testDescription, testFolder,
                         config, flavour, rendererParams)

# ------------------------------------------------------------------------------
class PodTester(Tester):
    def __init__(self, testPlan):
        Tester.__init__(self, testPlan, [], PodTestFactory)

# ------------------------------------------------------------------------------
if __name__ == '__main__': PodTester('Tests.odt').run()
# ------------------------------------------------------------------------------
